#!/usr/bin/env python

import logging
import platform
import sys

from utah.url import url_argument
from utah.group import check_user_group, print_group_error_message
from utah.client.runner import Runner
from utah.client.state_agent import StateAgentYAML
from utah.client.result import Result, ResultYAML, ResultJSON

from utah.client.common import DEFAULT_STATE_FILE
from utah.client.exceptions import UTAHClientError


def get_parser():
    """
    Process the command line arguments.
    """
    # requires Python2.7+
    import argparse

    parser = argparse.ArgumentParser(
        description='''Run test cases locally and write results to a file.

Note: This is expected to be used only for development purposes on a machine
that has been provisioned using run_utah_tests.py''',
        formatter_class=argparse.RawTextHelpFormatter)
    parser.add_argument('--resume', action='store_true',
                        help='Continue a previous run.  Used after a reboot')
    parser.add_argument('-s', '--state-file',
                        help=('File to use for storing state (default "%s")'
                              % DEFAULT_STATE_FILE))
    parser.add_argument('-f', '--format',
                        choices=['text', 'yaml', 'json'],
                        default='yaml',
                        help='Output format (default "%(default)s")')
    parser.add_argument('-t', '--testdir',
                        default='/var/lib/utah', help='Main test directory')
    parser.add_argument('-i', '--install-type',
                        choices=['desktop', 'server', 'mini', 'alternate'],
                        default='desktop',
                        help=('Installation Variant '
                              '(i.e. server, desktop, etc.)'))
    parser.add_argument('-r', '--runlist', type=url_argument,
                        # Having a default here means that if a resume happens
                        # post-reboot without a runlist specified, it will
                        # run the default runlist instead of erroring
                        #default='/usr/share/utah/client/examples/master.run',
                        help='runlist file name')
    parser.add_argument('-o', '--output', help='write output to this file')
    parser.add_argument('-a', '--append', action='store_true',
                        help='append to output')
    parser.add_argument('-d', '--debug', action='store_true',
                        help='Print debugging output')

    return parser


def exit(returncode, msg):
    print msg
    sys.exit(returncode)


def setup_logging(debug=False):
    if debug:
        level = logging.DEBUG
    else:
        level = logging.WARNING

    logging.basicConfig(level=level,
                        format='%(asctime)s - %(levelname)s - %(message)s')


def main():
    """
    The main driver for the 'utah' application.
    """
    # TODO: write <2.7 optparse version and set based on version of python
    #       being used.
    parser = get_parser()
    args = parser.parse_args()

    # When resuming, this script is invoked as root by rc.local
    # No group membership is assigned then, so root needs to be able to run
    # this script anyway
    if not check_user_group(rootok=args.resume):
        print_group_error_message(__file__)
        sys.exit(3)

    setup_logging(args.debug)

    logging.debug(args)
    logging.info(platform.uname())

    """
    print args
    import sys
    sys.exit(0)
    """
    resume = args.resume
    state_file = args.state_file
    testdir = args.testdir
    runlist = args.runlist
    output = args.output
    install_type = args.install_type

    if args.append:
        open_flags = 'a'
    else:
        open_flags = 'w'

    # Redirect stdout to a file if requested.
    if output:
        try:
            sys.stdout = open(output, open_flags)
        except IOError as e:
            exit(e.errno, e)

    if args.format == 'text':
        result_class = Result
    elif args.format == 'json':
        result_class = ResultJSON
    else:
        result_class = ResultYAML

    try:
        state_agent = StateAgentYAML(state_file=state_file)
        runner = Runner(install_type=install_type, state_agent=state_agent,
                        result_class=result_class, runlist=runlist,
                        resume=resume, testdir=testdir)
        if resume:
            runner.load_state()

        returncode = runner.run()

    except UTAHClientError as e:
        errno = getattr(e, 'errno', 1)
        exit(errno, e)

    sys.exit(returncode)

if __name__ == "__main__":
    main()
