#!/usr/bin/python

import argparse
import sys

from utah.exceptions import UTAHException
from utah.group import check_user_group, print_group_error_message
from utah.provisioning.inventory.sqlite import ManualCobblerSQLiteInventory
from utah.run import run_tests
from utah.url import url_argument


def get_parser():
    parser = argparse.ArgumentParser(
            description=('Provision a machine using cobbler '
                         'and run one or more UTAH runlists there.'),
            epilog=("For example:\n"
                    "Provision a machine using a precise server image "
                    "with i386 architecture and run the two given runlists\n"
                    "\t%(prog)s -s precise -t server -a i386 \\\n"
                    "\t\t/usr/share/utah/client/examples/master.run \\\n"
                    "\t\t'http://people.canonical.com/~max/max_test.run'"),
            formatter_class=argparse.RawDescriptionHelpFormatter)
    parser.add_argument('runlists', metavar='runlist', nargs='+',
                        help='URLs of runlist files to run')
    parser.add_argument('-n', '--name', help='Name of machine to provision')
    parser.add_argument('-i', '--image', type=url_argument,
                        help='Image/ISO file to use for installation')
    parser.add_argument('-p', '--preseed', type=url_argument,
                        help='Preseed file to use for installation')
    parser.add_argument('-s', '--series', metavar='SERIES',
                        choices=('hardy', 'lucid', 'natty',
                                 'oneiric', 'precise', 'quantal'),
                        help='Series to use for VM creation (%(choices)s)')
    parser.add_argument('-t', '--type', metavar='TYPE',
                        choices=('desktop', 'server', 'mini', 'alternate'),
                        help=('Install type to use for VM creation '
                              '(%(choices)s)'))
    parser.add_argument('-a', '--arch', metavar='ARCH',
                        choices=('i386', 'amd64'),
                        help=('Architecture to use for VM creation '
                              '(%(choices)s)'))
    parser.add_argument('-d', '--debug', action='store_true',
                        help='Enable debug logging')
    parser.add_argument('-j', '--json', action='store_true',
                        help='Enable json logging (default is YAML)')
    return parser


def run_test_cobbler(args=None):
    if not check_user_group():
        print_group_error_message(__file__)
        sys.exit(3)

    if args is None:
        args = get_parser().parse_args()

    locallogs = []
    exitstatus = 0
    machine = None

    for arg in ['arch', 'series', 'type']:
        if vars(args)[arg] is not None:
            sys.stderr.write('Automatic image fetch is not yet supported')
            sys.stderr.write('Please specify an image using the -i option')
            sys.stderr.write('Local files and http paths are supported')
            sys.exit(4)

    try:
        inventory = ManualCobblerSQLiteInventory()
        kw = {}
        for arg in ['image', 'preseed']:
            value = vars(args)[arg]
            if value is not None:
                kw.update([(arg, value)])
        machine = inventory.request(debug=args.debug, new=True,
                                    dlpercentincrement=10, **kw)
        exitstatus, locallogs = run_tests(args, machine)

    except UTAHException as error:
        mesg = 'Exception: ' + str(error)
        try:
            machine.logger.error(mesg)
        except (AttributeError, NameError):
            sys.stderr.write(mesg)
        exitstatus = 2

    finally:
        if machine is not None:
            try:
                machine.destroy()
            except UTAHException as error:
                sys.stderr.write('Failed to destroy machine: ' + str(error))

        if len(locallogs) != 0:
            print('Test logs copied to the following files:')
            print("\t" + "\n\t".join(locallogs))

    sys.exit(exitstatus)


if __name__ == '__main__':
    run_test_cobbler()
