#!/usr/bin/python

import argparse
import sys

from utah.exceptions import UTAHException
from utah.group import check_user_group, print_group_error_message
from utah.provisioning.inventory.sqlite import TinySQLiteInventory
from utah.run import run_tests


def get_parser():
    parser = argparse.ArgumentParser(
            description=('Create a virtual machine '
                         'and run a UTAH runlist there.'),
            epilog=("For example:\n"
                    "Provision a VM using a precise server image "
                    "with i386 architecture and run the two given runlists\n"
                    "\t%(prog)s -s precise -t server -a i386 \\\n"
                    "\t\t/usr/share/utah/client/examples/master.run \\\n"
                    "\t\t'http://people.canonical.com/~max/max_test.run'"),
            formatter_class=argparse.RawDescriptionHelpFormatter)
    parser.add_argument('runlists', metavar='runlist', nargs='+',
                        help='URLs of runlist files to run')
    parser.add_argument('-s', '--series', metavar='SERIES',
                        choices=('hardy', 'lucid', 'natty',
                                 'oneiric', 'precise', 'quantal'),
                        help='Series to use for VM creation (%(choices)s)')
    parser.add_argument('-t', '--type', metavar='TYPE',
                        choices=('desktop', 'server', 'mini', 'alternate'),
                        help=('Install type to use for VM creation '
                              '(%(choices)s)'))
    parser.add_argument('-a', '--arch', metavar='ARCH',
                        choices=('i386', 'amd64'),
                        help=('Architecture to use for VM creation '
                              '(%(choices)s)'))
    parser.add_argument('-n', '--no-destroy', action='store_true',
                        help='Preserve VM after tests have run')
    parser.add_argument('-d', '--debug', action='store_true',
                        help='Enable debug logging')
    parser.add_argument('-j', '--json', action='store_true',
                        help='Enable json logging (default is YAML)')
    return parser


def run_test_vm(args=None):
    if not check_user_group():
        print_group_error_message(__file__)
        sys.exit(3)

    if args is None:
        args = get_parser().parse_args()

    locallogs = []
    exitstatus = 0
    machine = None

    try:
        inventory = TinySQLiteInventory()
        kw = {}
        for arg in ['arch', 'series']:
            value = vars(args)[arg]
            if value is not None:
                kw.update([(arg, value)])
        if args.type is not None:
            kw.update([('installtype', args.type)])
        machine = inventory.request(debug=args.debug, new=True,
                                    dlpercentincrement=10, **kw)
        exitstatus, locallogs = run_tests(args, machine)

    except UTAHException as error:
        sys.stderr.write('Exception: ' + str(error))
        exitstatus = 2

    finally:
        if not args.no_destroy and machine is not None:
            try:
                machine.destroy()
            except UTAHException as error:
                sys.stderr.write('Failed to destroy machine: ' + str(error))
            finally:
                try:
                    inventory.destroy(machine.machineid)
                except UTAHException as error:
                    sys.stderr.write('Failed to update inventory: '
                                     + str(error))
                finally:
                    del machine
        if len(locallogs) != 0:
            print('Test logs copied to the following files:')
            print("\t" + "\n\t".join(locallogs))

        sys.exit(exitstatus)


if __name__ == '__main__':
    run_test_vm()
