"""
Common methods.
"""

import datetime
import re
import grp
import getpass
import json
import os
import platform
import pwd
import signal
import subprocess
import yaml
import jsonschema

from utah.client.exceptions import YAMLParsingError


PASS = 0
FAIL = 1
ERROR_TIMEOUT = -9

CONFIG = {
    'DEBUG': False,
    'TEST_DIR': '/var/lib/utah',
    'UTAH_CLIENT': '/usr/share/utah/client',
}

# Default TestSuite filenames
DEFAULT_TSLIST = 'tslist.run'
DEFAULT_TSCONTROL = 'ts_control'

UTAH_DIR = CONFIG['TEST_DIR']
UTAH_TESTDIR = os.path.join(UTAH_DIR, 'testsuites')
MASTER_RUNLIST = os.path.join(UTAH_DIR, "master.run")
UTAH_CLIENT = CONFIG['UTAH_CLIENT']
CLIENT_CONFIG = os.path.join(UTAH_DIR, 'config', 'client.json')
DEFAULT_STATE_FILE = os.path.join(UTAH_DIR, "state.yaml")

RETURN_CODES = {'PASS': 0, 'ERROR': -1, 'FAIL': 1}

CMD_TC_BUILD = 'testcase_build'
CMD_TC_SETUP = 'testcase_setup'
CMD_TC_TEST = 'testcase_test'
CMD_TC_CLEANUP = 'testcase_cleanup'

CMD_TS_FETCH = 'testsuite_fetch'
CMD_TS_BUILD = 'testsuite_build'
CMD_TS_SETUP = 'testsuite_setup'
CMD_TS_CLEANUP = 'testsuite_cleanup'


def do_nothing(obj=None):
    """
    Do nothing method used a placeholder for save_state_callbacks.
    """
    pass


# Inspired by:
# http://stackoverflow.com/a/3326559
def run_cmd(command, cwd=None, timeout=0, cmd_type=CMD_TC_TEST, run_as=None):

    if run_as is not None:
        # add -n so sudo will not prompt for a password on the tty
        cmd_prefix = "sudo -n -u {} ".format(run_as)
    else:
        cmd_prefix = ""

        # just a guess, but we're only displaying it
        run_as = getpass.getuser()

    if command is None:
        return

    class TimeoutAlarm(Exception):
        pass

    def alarm_handler(signum, frame):
        raise TimeoutAlarm

    start_time = datetime.datetime.now()
    p = subprocess.Popen("{} {}".format(cmd_prefix, command),
                         shell=True, cwd=cwd,
                         stdout=subprocess.PIPE, stderr=subprocess.PIPE)

    if timeout != 0:
        signal.signal(signal.SIGALRM, alarm_handler)
        signal.alarm(timeout)

    try:
        stdout, stderr = p.communicate()
        if timeout != 0:
            signal.alarm(0)
    except TimeoutAlarm:
        pids = [p.pid]
        # Kill p's children too.
        pids.extend(get_process_children(p.pid))

        for pid in pids:
            # process might have died before getting to this line
            # so wrap to avoid OSError: no such process
            try:
                os.kill(pid, signal.SIGKILL)
            except OSError:
                pass

        time_delta = datetime.datetime.now() - start_time

        return make_result(command=command,
                           retcode=ERROR_TIMEOUT,
                           start_time=str(start_time),
                           time_delta=str(time_delta),
                           cmd_type=cmd_type,
                           user=run_as,
                          )

    time_delta = datetime.datetime.now() - start_time

    return make_result(command=command,
                       retcode=p.returncode,
                       stdout=stdout,
                       stderr=stderr,
                       start_time=str(start_time),
                       time_delta=str(time_delta),
                       cmd_type=cmd_type,
                       user=run_as,
                      )


# TODO: it might make sense to have a result object that keeps track of
# test pass, fail, error and serializes it's data.
def make_result(command, retcode, stdout='', stderr='', start_time='',
                time_delta='', cmd_type=CMD_TC_TEST, user="unknown"):
    """
    Make a result dictionary.
    """
    res = {
        'command': command,
        'returncode': retcode,
        'stdout': stdout,
        'stderr': stderr,
        'start_time': start_time,
        'time_delta': time_delta,
        'cmd_type': cmd_type,
        'user': user,
    }

    return res


def get_process_children(pid):
    p = subprocess.Popen('ps --no-headers -o pid --ppid %d' % pid, shell=True,
                         stdout=subprocess.PIPE, stderr=subprocess.PIPE)
    stdout, stderr = p.communicate()

    return [int(pid) for pid in stdout.split()]


def parse_yaml_file(filename):
    """
    Parse yaml file and raise exception with error location
    in case of error
    """
    try:
        with open(filename, 'r') as fp:
            data = yaml.load(fp)
    except yaml.YAMLError as exception:
        if hasattr(exception, 'problem_mark'):
            mark = exception.problem_mark
            error_message = ('YAML parsing error '
                             'at line {line}, column {column}: {filename}'
                             .format(line=mark.line + 1,
                                     column=mark.column + 1,
                                     filename=filename))
        else:
            error_message = ('YAML parsing error: {filename}'
                             .format(filename=filename))
        raise YAMLParsingError(error_message)
    return data


def parse_control_file(filename, schema):
    """
    Parse a control file and check against the schema
    """
    control_data = parse_yaml_file(filename)
    jsonschema.validate(control_data, schema)
    return control_data


def debug_print(data, force=False):
    """
    Print debugging information if CONFIG['DEBUG'] is defined and True
    """
    try:
        debug = CONFIG['DEBUG']
    except NameError:
        debug = False

    debug = debug or force

    if debug:
        print "DEBUG:", data

    # return True if we printed some output
    return debug


def raise_privileges(orig_privs):
    """
    Raise privileges to original effective privileges.
    """
    os.seteuid(orig_privs['euid'])
    os.setegid(orig_privs['egid'])
    os.setgroups(orig_privs['groups'])


def drop_privileges(user='nobody', group='nogroup', full=False):
    """
    Change user and/or group to decrease privileges.

    Currently only drops effective privileges.

    Returns a dictionary of the original values.
    """

    orig_privs = {}

    orig_privs['uid'] = os.getuid()
    orig_privs['euid'] = os.geteuid()
    orig_privs['gid'] = os.getgid()
    orig_privs['egid'] = os.getegid()
    orig_privs['groups'] = os.getgroups()

    # If root, drop privileges
    # inspired by: http://stackoverflow.com/a/2699996
    if os.getuid() == 0 or os.geteuid() == 0:

        running_uid = pwd.getpwnam(user).pw_uid
        running_gid = grp.getgrnam(group).gr_gid
        os.setgroups([])

        if full:
            os.setgid(running_gid)
        else:
            os.setegid(running_gid)

        if full:
            os.setuid(running_uid)
        else:
            os.seteuid(running_uid)

        os.umask(077)

    return orig_privs


def gather_artifacts(extras=[], excludes=[]):
    """
    Include a set of useful artifacts (i.e. files) for the logs.
    """
    artifacts = ['/var/log/syslog', '/proc/cpuinfo']

    artifacts += extras

    # if excludes is a single item make it a list
    if not isinstance(excludes, list):
        excludes = [excludes]

    for exc in excludes:
        try:
            artifacts.remove(exc)
        except ValueError:
            print("{} is not already an artifact".format(exc))
            pass

    for artifact in artifacts:
        print("### Start {} ###".format(artifact))
        with open(artifact, 'r') as fp:
            for line in fp:
                print(line.strip())
        fp.close()
        print("### End {} ###".format(artifact))


def get_media_info():
    """
    Get the contents of the media-info file if available.
    """

    filename = '/var/log/installer/media-info'

    media_info = None

    if os.path.exists(filename):
        media_info = open(filename, 'r').read()

    return media_info


def get_host_info():
    """
    Get host info, useful for debugging.
    """

    retval = {}

    retval['uname'] = platform.uname()
    retval['media-info'] = get_media_info()

    return retval


def get_arch():
    """
    The host's architecture (i.e. i386, amd64, etc.)
    """
    uname = platform.uname()

    arch = {
            'x86_64': 'amd64',
            'x86': 'i386',
            }.get(uname[5])  # 5 == processor

    return arch


def get_release():
    """
    The host's release name (i.e. precise, quantal, etc.)
    """
    return platform.dist()[2]


def get_api_config():
    """
    Parse the client configuration file for API configuration data.
    """

    data = {}
    data['API'] = {}

    with open(CLIENT_CONFIG, 'r') as fp:
        data = json.load(fp)

    return data['API']


def get_build_number():
    host_info = get_host_info()
    pattern = re.compile('.*\(([0-9.]+)\)$')
    match = pattern.match(host_info['media-info'])

    return match.group(1) if match else '?'
