#!/usr/bin/env python
# phoenix.py - A bootstrap script for testsuite/testcase authors

import argparse
import os

TC_CONTROL_NAME = 'tc_control'
TS_CONTROL_NAME = 'ts_control'
TSLIST_NAME = 'tslist.run'
MASTER_NAME = 'master.run'
CHANGEME = "CHANGE_ME"


class Phoenix(object):

    def __init__(self, testsuite, directory='.', testcases=['testcase1']):
        self.testsuite = testsuite
        self.directory = directory
        self.testcases = testcases

        self.tslist_content = ""
        self.ts_control_content = """
build_cmd: echo building
"""
        self.tc_control_content = """
description: {0}
command: {0}
dependencies: {0}
action: |
  1. {0}
expected_results: |
  1. {0}
type: userland
timeout: 60
#build_cmd: {0}
#tc_setup: {0}
#tc_cleanup: {0}
""".format(CHANGEME)
        self.master_run_content = """
---
testsuites:
"""

        self.FILE_CONTENT = {
                TSLIST_NAME: self.tslist_content,
                TS_CONTROL_NAME: self.ts_control_content,
                TC_CONTROL_NAME: self.tc_control_content,
                MASTER_NAME: self.master_run_content,
                }

        self.FILES = [
                TSLIST_NAME,
                TS_CONTROL_NAME,
                MASTER_NAME,
                ]
        # Add testcase files
        for d in self.testcases:
            self.FILES.append(os.path.join(d, TC_CONTROL_NAME))
            self.FILE_CONTENT[TSLIST_NAME] += "- test: {name}\n".format(name=d)

        self.FILE_CONTENT[MASTER_NAME] += "  - name: {}\n".format(testsuite)
        self.FILE_CONTENT[MASTER_NAME] += \
                "    fetch_cmd: {}\n".format(CHANGEME)

        self.debug()

    def debug(self):
        print("DEBUG: FILES: {}".format(self.FILES))

        for f in self.FILE_CONTENT.keys():
            fc = self.FILE_CONTENT[f]
            print("DEBUG: {}: {}".format(f, fc))

    def create_file(self, path, content):
        if not os.path.exists(path):
            with open(path, 'w') as fp:
                fp.write(content)
                fp.close()

    def build_suite(self):
        if not os.path.exists(self.directory):
            os.mkdir(self.directory)

        # create if testsuite directory if it doesn't exist
        ts_dir = os.path.join(self.directory, self.testsuite)
        if not os.path.exists(ts_dir):
            os.mkdir(ts_dir)

        # create the ts_control file
        ts_file = os.path.join(ts_dir, TS_CONTROL_NAME)
        self.create_file(ts_file, self.FILE_CONTENT[TS_CONTROL_NAME])

        # create the master.run file
        master_file = os.path.join(ts_dir, MASTER_NAME)
        self.create_file(master_file, self.FILE_CONTENT[MASTER_NAME])

        # create the tslist.run file
        tslist_file = os.path.join(ts_dir, TSLIST_NAME)
        self.create_file(tslist_file, self.FILE_CONTENT[TSLIST_NAME])

        # create the test case directories
        for tc in self.testcases:
            tc_dir = os.path.join(ts_dir, tc)
            tc_file = os.path.join(tc_dir, TC_CONTROL_NAME)

            # Don't create the directory if it exists
            if not os.path.exists(tc_dir):
                os.mkdir(tc_dir)

            self.create_file(tc_file, self.FILE_CONTENT[TC_CONTROL_NAME])


def main():

    parser = argparse.ArgumentParser(description='phoenix bootstrapper')
    parser.add_argument('testsuite', metavar='TESTSUITE', type=str, nargs=1,
            help='name of the testsuite to create')
    parser.add_argument('testcases', metavar='TESTCASE', type=str, nargs='*',
            help='names of testcases to create')
    parser.add_argument('--directory', '-D', dest='directory', default='.',
            help='where to create the testsuite')

    args = parser.parse_args()

    phoenix = Phoenix(args.testsuite[0],
            directory=args.directory,
            testcases=args.testcases,
            )
    phoenix.build_suite()

if __name__ == "__main__":
    main()
