from common import run_cmd, DEFAULT_TSLIST
from common import CMD_TS_FETCH
from result import Result
from testsuite import TestSuite
from state_agent import StateAgentYAML
import exceptions

import os
import shutil
import stat
import urllib
import jsonschema

from common import (
        MASTER_RUNLIST,
        UTAH_DIR,
        DEFAULT_STATE_FILE,
        RETURN_CODES,
        parse_yaml_file,
        )

RC_LOCAL = '/etc/rc.local'
RC_LOCAL_BACKUP = '%s-utah.bak' % RC_LOCAL

# XXX: need to output to the same file that was supplied on the original
# run.
rc_local_content = """#!/bin/sh

/usr/bin/utah --resume --append -o /var/lib/utah/utah.out -r {runlist}
"""


class Runner(object):
    """
    The main runner.

    Parses a master runlist, builds a list of TestSuites, and runs them.
    """

    status = "NOTRUN"
    utah_exec_path = '/usr/bin/utah'

    TESTSUITE_ENTRY_SCHEMA = {
        'type': 'object',
        'properties': {
            'name': {
                'type': 'string',
                'required': True,
                },
            'fetch_cmd': {
                'type': 'string',
                'required': True,
                },
            'include_tests': {
                'type': 'array',
                'items': {'type': 'string'},
                },
            'exclude_tests': {
                'type': 'array',
                'items': {'type': 'string'},
                },
        },
    }

    TESTSUITE_INCLUDE_SCHEMA = {
        'type': 'object',
        'properties': {
            'include': {
                'type': 'string',
                'required': True,
            },
        },
    }

    MASTER_RUNLIST_SCHEMA_ORIG = {
        'type': 'array',
        'items': {
            'type': [TESTSUITE_ENTRY_SCHEMA, TESTSUITE_INCLUDE_SCHEMA],
            'required': True,
            },
    }

    MASTER_RUNLIST_SCHEMA_NEW = {
        'type': 'object',
        'properties': {
            "timeout": {
                "type": "integer",
                "minimum": 0,
                },
            "repeat_count": {
                "type": "integer",
                "minimum": 0,
                },
            "type": {
                'type': 'string',
                'enum': ['smoke', 'kernel-sru', 'bootspeed', 'upgrade'],
                },
            "publish": {
                'type': 'object',
                'properties': {
                    "url": {
                        "type": "string",
                        "required": True,
                        },
                    "token": {
                        "type": "string",
                        "required": True,
                        },
                    "name": {
                        "type": "string",
                        "required": True,
                        },
                    },
                },
            "testsuites": {
                # must be a list to accept a schema rather than a simple type
                "type": [MASTER_RUNLIST_SCHEMA_ORIG],
                "required": True,
                },
            }
        }

    MASTER_RUNLIST_SCHEMA = {
            'type': [MASTER_RUNLIST_SCHEMA_ORIG, MASTER_RUNLIST_SCHEMA_NEW],
            'required': True,
            }

    def __init__(self, install_type, runlist=None, result_class=Result,
            testdir=UTAH_DIR, state_agent=None, resume=False):

        # Runlist URL passed through the command line
        self.testdir = testdir

        self.testsuitedir = os.path.join(testdir, 'testsuites')

        # The testdir must exist, XXX: perhaps we can create it, needs
        # discussion.
        if not os.path.exists(testdir):
            raise exceptions.BadDir(testdir)

        # create the testsuite directory
        if not os.path.exists(self.testsuitedir):
            os.mkdir(self.testsuitedir)

        # If no runlist is supplied look for one in the testdir
        if runlist is None:
            runlist = os.path.join(testdir, 'master.run')

        self.master_runlist = runlist
        self.suites = []
        self.result = result_class()
        self.result_class = result_class
        self.state_file = DEFAULT_STATE_FILE
        self.fetched_suites = {}
        self.timeout = None
        self.install_type = install_type

        self.errors = 0
        self.passes = 0
        self.failures = 0
        self.fetch_errors = 0

        self.state_agent = state_agent or StateAgentYAML()

        # Cleanup the state file if this is supposed to be a fresh run.
        if not resume:
            self.state_agent.clean()
        else:
            self.reset_rc_local()
            self.load_state()

        try:
            os.chdir(self.testsuitedir)
        except OSError as e:
            raise exceptions.BadDir(str(e))

        # needs to be run prior to process_master_runlist since that's where
        # test suites are fetched.
        self.fetched_suites = self.get_fetched_suites()

        self.process_master_runlist(resume=resume)
        self.result.runlist = self.master_runlist
        self.result.install_type = self.install_type

    def backup_rc_local(self):
        """
        Backup /etc/rc.local if it exists.
        """

        # Ignore permission denied errors since we only care if a reboot is
        # pending whether or not we can write to RC_LOCAL(_BACKUP).
        try:
            os.rename(RC_LOCAL, RC_LOCAL_BACKUP)
        except OSError as e:
            if e.errno in [13]:
                pass

    def reset_rc_local(self):
        """
        Restore original /etc/rc.local if there is a backup
        Otherwise remove it.
        """

        # Ignore permission denied errors since we only care if a reboot is
        # pending whether or not we can write to RC_LOCAL(_BACKUP).
        try:
            if os.path.exists(RC_LOCAL_BACKUP):
                os.rename(RC_LOCAL_BACKUP, RC_LOCAL)
            elif os.path.exists(RC_LOCAL):
                os.remove(RC_LOCAL)
        except OSError as e:
            if e.errno in [13]:
                pass

    def setup_rc_local(self, rc_local=RC_LOCAL, runlist=None):
        """
        Setup /etc/rc.local to kick-off a --resume on successful boot.
        """
        runlist = runlist or self.master_runlist or MASTER_RUNLIST

        self.rc_local_content = rc_local_content

        fp = open(rc_local, 'w')
        fp.write(self.rc_local_content.format(runlist=runlist))
        fp.close()

        os.chmod(rc_local, stat.S_IRWXU | stat.S_IROTH | stat.S_IRGRP)

    def process_results(self):
        # Add stats to the results
        self.result.failures = self.failures
        self.result.passes = self.passes
        self.result.errors = self.errors
        self.result.fetch_errors = self.fetch_errors

        # process the results
        self.result.result()
        return self.returncode()

    def run(self):
        """
        Run the test suites we've parsed.
        """

        # Return value to indicate whether processing of a Runner should
        # continue.  This is to avoid a shutdown race on reboot cases.
        keep_going = True

        self.status = "RUN"
        for suite in self.suites:
            # Start in self.testdir.
            os.chdir(self.testsuitedir)

            if not suite.is_done():
                keep_going = suite.run()
                if not keep_going:
                    return self.process_results()

                self.errors += suite.errors
                self.passes += suite.passes
                self.failures += suite.failures

        self.status = "DONE"
        self.save_state()

        return self.process_results()

    def add_suite(self, suite):
        self.suites.append(suite)

    def get_fetched_suites(self):
        """
        Return a list of fetched suites from the state_agent.
        """
        state = self.state_agent.load_state()

        fetched_suites = []

        if state and 'fetched_suites' in state:
            fetched_suites = state['fetched_suites']

        return fetched_suites

    def load_state(self):
        state = self.state_agent.load_state()

        self.master_runlist = state['master_runlist']
        self.status = state['status']
        self.result_class = state['result_class']

        self.suites = []
        for state_suite in state['suites']:
            suite = TestSuite(name=state_suite['name'],
                              path=self.testsuitedir,
                              result=self.result,
                              timeout=self.timeout,
                              _save_state_callback=self.save_state,
                              _reboot_callback=self.reboot)
            suite.load_state(state_suite)
            self.suites.append(suite)

        self.fetched_suites = state['fetched_suites']

    def save_state(self):
        """
        Save the list of tests we are to run and whether we've run them.
        """

        self.backup_runlist = os.path.join(self.testdir, 'master.run-reboot')

        if (os.path.exists(self.master_runlist)
            and self.master_runlist != self.backup_runlist):
            shutil.copyfile(self.master_runlist, self.backup_runlist)

        state = {
            'master_runlist': self.backup_runlist,
            'original_runlist': self.master_runlist,
            'status': self.status,
            'passes': self.passes,
            'failures': self.failures,
            'errors': self.errors,
            'fetch_errors': self.fetch_errors,
            'result_class': self.result_class,
            'suites': [],
            'fetched_suites': self.fetched_suites,
            'testdir': self.testdir
        }

        for suite in self.suites:
            state['suites'].append(suite.save_state())

        # hand the state dictionary off to the StateAgent to process
        self.state_agent.save_state(state)

        return state

    def count_suites(self):
        return len(self.suites)

    def count_tests(self):
        tests = 0

        for suite in self.suites:
            tests += suite.count_tests()

        return tests

    def process_master_runlist(self, runlist=None, resume=False):
        """
        Parse a master runlist building a list of suites from the parsed data.

        """
        runlist = runlist or self.master_runlist

        # Download runlist using the URL passed through the commmand line to
        # local filename if needed (it might be a local file already or cached)
        try:
            local_filename = urllib.urlretrieve(runlist)[0]
        except IOError:
            raise exceptions.MissingFile(runlist)

        data = parse_yaml_file(local_filename)
        try:
            jsonschema.validate(data, self.MASTER_RUNLIST_SCHEMA)
        except jsonschema.ValidationError as exception:
            raise exceptions.ValidationError(
                    'Master runlist failed to validate: {!r}\n'
                    'Detailed information: {}'
                    .format(local_filename, exception))

        if 'timeout' in data:
            self.timeout = int(data['timeout'])

        # Add publish metatdata to the result object
        if 'publish' in data:
            self.result.publish = data['publish']

        seen = []

        orig_dir = os.getcwd()

        if 'testsuites' in data:
            suites = data['testsuites']
        elif type(data) == list:
            suites = data
        else:
            raise exceptions.BadMasterRunlist(str(data))

        suites = data['testsuites'] if 'testsuites' in data else data
        for suite in suites:
            # Allow the inclusion of other master.run files
            if 'include' in suite:
                self.process_master_runlist(suite['include'])
                continue

            os.chdir(orig_dir)
            fetch_cmd = suite['fetch_cmd']

            name = suite['name']
            includes = suite.get('include_tests')
            excludes = suite.get('exclude_tests')

            suite_runlist = suite.get('runlist', DEFAULT_TSLIST)

            if name in seen:
                raise Exception("%s duplicated in runlist" % name)

            # Fetch the testsuite. On resume don't remove the testsuite
            # directory.
            if not resume and os.path.exists(name):
                shutil.rmtree(name)
            if not os.path.exists(name):
                os.mkdir(name)

            if name not in self.fetched_suites:
                self.result.add_result(run_cmd(fetch_cmd, cwd=name,
                                               cmd_type=CMD_TS_FETCH))
                self.fetched_suites.append(name)

            # If fetch_cmd failed move on to the next testsuite.
            if self.result.status != 'PASS':
                self.fetch_errors += 1
                continue

            # Create a TestSuite
            # TODO: pass master.run overrides into the test suite.
            s = TestSuite(name=name, runlist_file=suite_runlist,
                          includes=includes, excludes=excludes,
                          result=self.result, path=self.testsuitedir,
                          timeout=self.timeout,
                          _save_state_callback=self.save_state,
                          _reboot_callback=self.reboot)
            self.add_suite(s)

    def get_next_suite(self):
        """
        Return the next suite to be run.

        Mainly used for debugging.
        """

        suite = None

        for s in self.suites:
            if not s.is_done():
                suite = s
                break

        return suite

    def get_next_test(self):
        """
        Return the next test to be run.

        Mainly used for debugging.
        """
        return self.get_next_suite().get_next_test()

    def reboot(self):
        """
        Reboot the machine.

        Save state, setup /etc/rc.local, and shutdown.
        """

        self.result.result()

        self.save_state()

        self.backup_rc_local()
        self.setup_rc_local(runlist=self.backup_runlist)

        import subprocess
        subprocess.call(['shutdown', '-r', 'now'])

        # End of execution

    def returncode(self):
        if self.errors > 0 or self.fetch_errors > 0:
            return RETURN_CODES['ERROR']
        elif self.failures > 0:
            return RETURN_CODES['FAIL']
        else:
            return RETURN_CODES['PASS']

    def report(self):
        tests_run = self.passes + self.errors + self.failures

        output = ("total: %d, passes: %d, failure: %d, error: %d"
                  % (tests_run, self.passes, self.failures,
                     self.errors + self.fetch_errors))

        if self.errors > 0 or self.fetch_errors > 0:
            result = "ERROR"
        elif self.failures > 0:
            result = "FAIL"
        else:
            result = "PASS"

        return "%s - %s" % (result, output)
