"""
Testcase specific code.
"""
import jsonschema

from common import run_cmd, parse_control_file
from common import do_nothing
from common import CMD_TC_BUILD, CMD_TC_SETUP, CMD_TC_TEST, CMD_TC_CLEANUP
from exceptions import MissingFile, ValidationError, MissingData


class TestCase(object):
    """
    The TestCase class.

    status is one of 'NOTRUN', 'BUILD', 'SETUP', 'RUN', 'CLEANUP', or 'DONE'
    """

    status = 'NOTRUN'
    build_cmd = None
    tc_setup = None
    tc_cleanup = None
    reboot = 'never'
    reboot_callback = None
    save_state_callback = do_nothing
    type = 'userland'

    CONTROL_SCHEMA = {
        'type': 'object',
        'properties': {
            'description': {
                'type': 'string',
                'required': True,
                },
            'dependencies': {
                'type': 'string',
                'required': True,
                },
            'action': {
                'type': 'string',
                'required': True,
                },
            'expected_results': {
                'type': 'string',
                'required': True,
                },
            'type': {
                'type': 'string',
                'enum': ['userland']
                },
            'timeout': {'type': 'integer'},
            'build_cmd': {'type': 'string'},
            'command': {
                'type': 'string',
                'required': True,
                },
            'tc_setup': {'type': 'string'},
            'tc_cleanup': {'type': 'string'},
            'reboot': {
                'type': 'string',
                'enum': ['always', 'pass', 'never'],
                },
            },
        }

    def __init__(self, name, path, result, command=None, timeout=None,
                 _control_data=None, _save_state_callback=None,
                 _reboot_callback=None):
        """
        Build a TestCase from a control file's data.

        'name' is a directory where the test case resides.
        """
        self.name = name
        self.path = path
        self.working_dir = path
        self.filename = "%s/tc_control" % path
        self.results = []
        self.result = result
        self.timeout = timeout
        self.command = command
        self.reboot_callback = _reboot_callback
        self.run_as = None

        # The following are testcase documentation and should all be provided
        # in a tc_control file or via the state file.
        self.description = None
        self.dependencies = None
        self.action = None
        self.expected_results = None

        if _save_state_callback is not None:
            self.save_state_callback = _save_state_callback

        control_data = None
        import os
        if _control_data is None:
            if os.path.exists(self.filename):
                try:
                    control_data = parse_control_file(self.filename,
                                                      self.CONTROL_SCHEMA)
                except jsonschema.ValidationError as exception:
                    raise ValidationError(
                            '{!r} test case control file invalid: {!r}\n'
                            'Detailed information: {}'
                            .format(self.name, self.filename, exception))
            else:
                raise MissingFile(self.filename)
        else:
            control_data = _control_data

        if control_data is not None:
            if self.timeout is not None:
                control_data['timeout'] = self.timeout

            for key, value in control_data.iteritems():
                setattr(self, key, value)

        # Set any values passed into __init__() as final values
        if timeout is None and self.timeout is None:
            self.timeout = 0

        if command is not None:
            self.command = command

    def __str__(self):
        return ("%s: %s, %s, %s"
                % (self.name, self.description, self.command, self.timeout))

    def set_status(self, status):
        """
        Set the status for the test case and call the save state callback.
        """
        self.status = status
        self.save_state_callback()

    def build(self, result):
        """
        Run build, but only if we haven't started a run yet.
        """
        if self.status == 'NOTRUN':
            self.set_status('BUILD')
            result.add_result(run_cmd(self.build_cmd, cwd=self.working_dir,
                                      cmd_type=CMD_TC_BUILD))

    def setup(self, result):
        """
        Run tc_setup, but only if build() has just passed.
        """
        if self.status == 'BUILD' and result.status == 'PASS':
            self.set_status('SETUP')
            result.add_result(run_cmd(self.tc_setup, cwd=self.working_dir,
                                      cmd_type=CMD_TC_SETUP))

    def cleanup(self, result):
        """
        Run tc_cleanup after a run.
        """
        if self.status == 'RUN':
            self.set_status('CLEANUP')
            result.add_result(run_cmd(self.tc_cleanup, cwd=self.working_dir,
                                      cmd_type=CMD_TC_CLEANUP))

    def run(self):
        """
        Run the test case; including any build, setup, and cleanup commands.
        """

        if self.is_done():
            self.run_status = 'PASS'
            return self.run_status

        # Do not proceed if we are missing data
        if (self.description is None
                or self.dependencies is None
                or self.action is None
                or self.expected_results is None):
                raise MissingData

        # Return value to indicate whether processing of a TestSuite should
        # continue.  This is to avoid a shutdown race on reboot cases.
        keep_going = True

        result = self.result

        result.testcase = self.name

        self.build(result)

        # if the build fails don't run the test command
        if result.status != 'PASS':
            self.run_status = result.status

            # reset status
            result.status = 'PASS'
            return result.status

        self.setup(result)

        # if the setup fails don't run the test command
        if result.status != 'PASS':
            self.run_status = result.status

            # reset status
            result.status = 'PASS'
            return result.status

        if self.status == 'SETUP':
            timeout = self.timeout or 0
            self.status = "RUN"
            self.save_state_callback()
            extra_info = {
                    'description': self.description,
                    'dependencies': self.dependencies,
                    'action': self.action,
                    'expected_results': self.expected_results,
                    }
            result.add_result(run_cmd(self.command,
                                      timeout=timeout,
                                      cwd=self.working_dir,
                                      cmd_type=CMD_TC_TEST,
                                      run_as=self.run_as,
                                      ),
                                      extra_info=extra_info)

        # Clean up whether 'command' failed or not.
        self.cleanup(result)

        need_reboot = False
        if (self.reboot == 'always'
            or self.reboot == 'pass'
            and result.status == 'PASS'):
            need_reboot = True

        self.run_status = result.status
        result.status = 'PASS'

        if self.status == 'CLEANUP':
            self.status = "DONE"
            self.save_state_callback()

        if need_reboot and self.reboot_callback is not None:
            self.reboot_callback()
            keep_going = False

        return keep_going

    def process_overrides(self, overrides):
        """
        Sets override values from a TestSuite runlist for this test case.
        """
        for key, value in overrides.iteritems():
            setattr(self, key, value)

    def load_state(self, state):
        """
        Restore state from the supplied dictionary.

        Requires that 'state' has the same fieldnames as the TestCase class.
        """
        for key, value in state.iteritems():
            setattr(self, key, value)

    def save_state(self):
        """
        Returns a dictionary representing the test's state.
        """
        state = {
                'name': self.name,
                'path': self.path,
                'command': self.command,
                'timeout': self.timeout,
                'status': self.status,
                'build_cmd': self.build_cmd,
                'tc_setup': self.tc_setup,
                'tc_cleanup': self.tc_cleanup,
                'type': self.type,
                'description': self.description,
                'dependencies': self.dependencies,
                'action': self.action,
                'expected_results': self.expected_results,
        }

        return state

    def is_done(self):
        """
        Determine if the case is done.  This might mean that something has
        failed.  Used by suite to determine if the suite needs to be re-run
        on resume.
        """
        return self.status == 'DONE' or self.status == 'CLEANUP'
