######################################################################
### Privileges Tests
######################################################################

import os
import pwd
import grp
import shlex
import subprocess
import unittest

from utah.client.common import drop_privileges, raise_privileges

NOBODY_UID = pwd.getpwnam('nobody').pw_uid
NOGROUP_GID = grp.getgrnam('nogroup').gr_gid

TMP_CONTENT = """Some data"""
TMP_FILENAME = '/tmp/priv_test.tmp'


def print_priv(msg=None):
    """
    Print out the current privileges.
    """
    if msg is not None:
        print(msg)

    print("uid: {}, euid: {}".format(os.getuid(), os.geteuid()))
    print("gid: {}, egid: {}".format(os.getgid(), os.getegid()))


def write_temp_file(filename=TMP_FILENAME):
    """
    Write the contents to a temp file.
    """
    with open(filename, 'w') as fp:
        fp.write(TMP_CONTENT)
    fp.close()


def read_temp_file(filename=TMP_FILENAME):
    """
    Print the contents of a temp file.
    """
    with open(filename, 'r') as fp:
        for line in fp:
            print(line.strip())
    fp.close()


def _do_cmd(cmd, quiet=False, shell=False):
    """
    Run a command.
    """
    cmd_args = cmd

    if not shell:
        cmd_args = shlex.split(cmd)

    proc = subprocess.Popen(
        cmd_args,
        shell=shell,
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE
    )

    (stdout, stderr) = proc.communicate()

    if not quiet:
        print("{}:\n{}".format(cmd, stdout))


def stat_file(filename=TMP_FILENAME):
    proc = subprocess.Popen(['ls', '-al', filename], stdout=subprocess.PIPE,
            stderr=subprocess.PIPE)

    (stdout, stderr) = proc.communicate()
    print("{}: {}".format(filename, 'stdout'))


def get_privs():
    privs = {}

    privs['uid'] = os.getuid()
    privs['euid'] = os.geteuid()
    privs['gid'] = os.getgid()
    privs['egid'] = os.getegid()

    return privs


class baseTestPriv(unittest.TestCase):

    def assertRoot(self, msg=None):
        self.assertEqual(os.geteuid(), 0, msg="Must be root")

    def assertPrivs(self, privs, msg=None):
        # Only check privileges passed in
        if 'uid' in privs:
            self.assertEqual(os.getuid(), privs['uid'], msg)

        if 'euid' in privs:
            self.assertEqual(os.geteuid(), privs['euid'], msg)

        if 'gid' in privs:
            self.assertEqual(os.getuid(), privs['uid'], msg)

        if 'egid' in privs:
            self.assertEqual(os.geteuid(), privs['euid'], msg)

    def setUp(self):
        print_priv("before")
        _do_cmd("id")

    def tearDown(self):
        print_priv("after")
        _do_cmd("id")


class TestPriv(baseTestPriv):
    def test_drop_priv(self):
        """
        Test that we drop effective privileges appropriately.
        """
        self.assertRoot()
        old_privs = drop_privileges()

        self.assertPrivs({'euid': NOBODY_UID, 'egid': NOGROUP_GID})

        print_priv("middle")
        _do_cmd("id")

        write_temp_file()
        read_temp_file()
        stat_file()

        raise_privileges(old_privs)

    def test_restore_priv(self):
        """
        Test that we restore effective privileges appropriately.
        """
        self.assertRoot()
        old_privs = drop_privileges()

        self.assertPrivs({'euid': NOBODY_UID, 'egid': NOGROUP_GID})

        print_priv("middle")

        raise_privileges(old_privs)

        self.assertPrivs(old_privs)


class TestXXXLast(baseTestPriv):
    """
    Test's to be run last.
    """
    def test_xxx_drop_priv_full(self):
        """
        Test dropping all privileges.
        """

        self.assertRoot()
        drop_privileges(full=True)
        self.assertPrivs({
            'uid': NOBODY_UID,
            'euid': NOBODY_UID,
            'gid': NOGROUP_GID,
            'egid': NOGROUP_GID,
            })

        _do_cmd("id")
