import jsonschema
import os
import re
import unittest
import yaml

from utah.client.runner import Runner

yaml_content_publish = """---
timeout: 80
publish:
    url: http://dashboard.local/api/add_result
    token: testtoken
    name: precise-server-amd64
testsuites:
    - name: testsuite1
      fetch_cmd: fetch
    - name: testsuite2
      fetch_cmd: fetch
"""

yaml_content_publish_bad = [
        re.sub('\n\s+url: .*', '', yaml_content_publish),
        re.sub('\n\s+token: .*', '', yaml_content_publish),
        re.sub('\n\s+name: .*', '', yaml_content_publish),
        ]

yaml_content = """---
- name: testsuite1
  fetch_cmd: get it now
- name: testsuite2
  fetch_cmd: fetch
"""

yaml_content_new = """---
timeout: 101
repeat_count: 99
testsuites:
    - name: testsuite1
      fetch_cmd: fetch1
    - name: testsuite2
      fetch_cmd: fetch2
"""
yaml_content_include = """---
testsuites:
    - include: master.run.other
    - name: testsuite1
      fetch_cmd: fetch1
    - name: testsuite2
      fetch_cmd: fetch2
"""

yaml_content_bad = [
"""---
testsuties:
    - name: testsuite1
      fetch_cmd: fetch1
""",  # typo
"""---
timeout: 101
repeat_count: 99
""",  # missing testsuites
"""---
- name: blah
- name: blip
""",  # missing fetch_cmds
"""---
testsuites:
  - name: blah
  - name: blip
""",  # missing fetch_cmds
"""---
testsuites:
""", # no testsuites
]


class TestJSONSchema(unittest.TestCase):

    def test_orig_schema(self):
        """
        Test that the original master.run is correctly validated by the schema.
        """

        data = yaml.load(yaml_content)
        print("data: {}".format(data))
        jsonschema.validate(data, Runner.MASTER_RUNLIST_SCHEMA)

    def test_multi_schema(self):
        """
        Test that the new master.run is correctly validated by the schema.
        """

        data_new = yaml.load(yaml_content_new)
        print("data_new: {}".format(data_new))
        jsonschema.validate(data_new, Runner.MASTER_RUNLIST_SCHEMA)

    def test_example_runlist(self):
        example_runlist = os.path.join(os.path.dirname(__file__),
                                       "../examples/master.run")

        self.assertTrue(os.path.exists(example_runlist))

        with open(example_runlist, "r") as fp:
            data = yaml.load(fp)
            print("data: {}".format(data))
            jsonschema.validate(data, Runner.MASTER_RUNLIST_SCHEMA)
            fp.close()

    def test_publish_schema(self):
        """
        Test that the publish piece of the schema works correctly.
        """

        data = yaml.load(yaml_content_publish)
        print("data: {}".format(data))
        jsonschema.validate(data, Runner.MASTER_RUNLIST_SCHEMA)

    def test_include_schema(self):
        """
        Test that include works correctly.
        """

        data = yaml.load(yaml_content_include)
        print("data: {}".format(data))
        jsonschema.validate(data, Runner.MASTER_RUNLIST_SCHEMA)

    def test_publish_bad_schema(self):
        """
        Test that the publish piece of the schema fails if any of
        the required field are missing.
        """

        for content in yaml_content_publish_bad:
            data = yaml.load(content)
            print("data: {}".format(data))
            self.assertRaises(jsonschema.ValidationError,
                              jsonschema.validate,
                              data, Runner.MASTER_RUNLIST_SCHEMA)

    def test_bad_schemas(self):
        """
        Test that required fields are enforced by the schema.
        """

        for content in yaml_content_bad:
            data = yaml.load(content)
            print("data: {}".format(data))
            self.assertRaises(jsonschema.ValidationError,
                              jsonschema.validate,
                              data, Runner.MASTER_RUNLIST_SCHEMA)
