import os
import shutil
import unittest

from utah.client.result import ResultYAML
from utah.client.state_agent import StateAgentYAML
from utah.client.runner import Runner
from utah.client.common import RETURN_CODES
from utah.client import exceptions

from .common import (  # NOQA
        setUp,  # Used by nosetests
        tearDown,  # Used by nosetests
        master_runlist_content,
        )


class TestRunner(unittest.TestCase):
    def setUp(self):
        self.result_class = ResultYAML
        self.state_file = '/tmp/state.yaml'
        self.state_agent = StateAgentYAML(state_file=self.state_file)
        self.runner = Runner(install_type='desktop',
                             result_class=self.result_class,
                             state_agent=self.state_agent)
        self.bad_dir = "/tmp/does_not_exist"

        self.assertFalse(os.path.exists(self.bad_dir),
                "{} shouldn't exist".format(self.bad_dir))

    def tearDown(self):

        # Remove self.bad_dir if it got created.
        try:
            shutil.rmtree(self.bad_dir)
        except OSError as e:
            # ignore 'No such file or directory errors
            if e.errno == 2:
                pass
            else:
                raise

    def test_fetch_failed(self):
        print "Starting test"
        # exit 2 should be flagged as an error.
        bad_runlist_content = ('---\n'
                               '- name: fake_tests\n'
                               '  fetch_cmd: exit 2\n')
        runlist = '/tmp/master_fetch_fails.run'

        self.assertFalse(os.path.exists(runlist), runlist)

        fp = open(runlist, 'w')
        fp.write(bad_runlist_content)
        fp.close()

        runner = Runner(install_type='desktop', result_class=self.result_class,
                state_agent=self.state_agent, runlist=runlist)

        # remove the runlist as early as possible to avoid leaving it
        # laying around if this test fails.
        try:
            os.remove(runlist)
        except OSError:
            pass

        runner.run()
        print runner.report()

        self.assertEqual(runner.fetch_errors, 1)

    def test_rc_local(self):
        tmp_rc_local = '/tmp/rc.local'
        self.runner.setup_rc_local(tmp_rc_local)

        self.assertTrue(os.path.exists(tmp_rc_local))

    def test_run(self):
        retcode = self.runner.run()

        for suite in self.runner.suites:
            self.assertTrue(suite.is_done())

        self.assertEqual(retcode, RETURN_CODES['ERROR'])

    def test_missing_testdir(self):
        """
        Test that the runner fails gracefully on missing testdir.
        """
        self.assertRaises(exceptions.BadDir,
                          Runner,
                          install_type="desktop",
                          result_class=self.result_class,
                          testdir=self.bad_dir)

    def test_missing_master_runlist(self):
        """
        Test that the runner fails gracefully on missing 'master.run'
        """

        tmp_master_runlist = os.path.join(self.bad_dir, 'master.run')

        if not os.path.exists(self.bad_dir):
            os.mkdir(self.bad_dir)

        # Make sure there isn't a master runlist
        if os.path.exists(tmp_master_runlist):
            os.remove(tmp_master_runlist)

        self.assertRaises(exceptions.MissingFile,
                          Runner,
                          install_type="desktop",
                          result_class=self.result_class,
                          runlist=tmp_master_runlist,
                          testdir=self.bad_dir)

    def test_runlist(self):
        """
        Test that passing a runlist to the Runner works properly.
        """
        runlist = '/tmp/master_runlist_test'

        fp = open(runlist, 'w')
        fp.write(master_runlist_content)
        fp.close()

        runner = Runner(install_type='desktop',
                        result_class=self.result_class,
                        runlist=runlist)
        runner.save_state()

        print("tests: {}".format(runner.count_tests()))

        self.assertTrue(runner.count_tests() > 0)

        os.remove(runlist)

    def test_testdir(self):
        """
        Test that the default test directory is created.
        """

        testdir = self.runner.testdir

        self.assertTrue(os.path.exists(testdir))
