import os
import shutil
import unittest
import yaml

from utah.client.state_agent import StateAgentYAML
from utah.client.runner import Runner
from utah.client.result import ResultYAML
from utah.client.common import UTAH_DIR

from .common import (  # NOQA
        setUp,  # Used by nosetests
        tearDown,  # Used by nosetests
        get_module_path,
        )

partial_state_file_content = """master_runlist: /var/lib/utah/master.run
status: RUN
result_class: !!python/name:utah.client.result.ResultYAML ''
fetched_suites:
- utah_tests
- utah_tests_sample
suites:
- build_cmd: null
  name: utah_tests
  status: INPROGRESS
  tests:
    - build_cmd: echo "building for test_one"
      command: python test_one.py
      description: A first sample test case.
      name: test_one
      path: utah_tests/test_one
      status: DONE
      tc_cleanup: echo "cleanup for test_one"
      tc_setup: echo "setup for test_one"
      timeout: 3
      type: userland
    - build_cmd: echo "building for test_two"
      command: python test_two.py
      description: A second sample test case.
      name: test_two
      path: utah_tests/test_two
      status: SETUP
      tc_cleanup: echo "cleanup for test_two"
      tc_setup: echo "setup for test_two"
      timeout: 300
      type: userland
  timeout: 0
  ts_cleanup: null
  ts_setup: null
- build_cmd: null
  name: utah_tests_sample
  status: DONE
  tests:
    - build_cmd: echo "building for sample_one"
      command: python sample.py
      description: A sample test case.
      name: sample_one
      path: utah_tests_sample/sample_one
      status: DONE
      tc_cleanup: echo "cleanup for sample_one"
      tc_setup: echo "setup for sample_one"
      timeout: 200
      type: userland
  timeout: 0
  ts_cleanup: null
  ts_setup: null
"""


class TestStateAgentYAML(unittest.TestCase):
    def setUp(self):
        """
        Create a state agent for testing.
        """
        self.state_file = '/tmp/state.yaml'

        # Fail if there is already a test state file.
        self.assertFalse(os.path.exists(self.state_file),
                         "state file (%s) already exists" % self.state_file)

        self.state_agent = StateAgentYAML(state_file=self.state_file)
        self.runner = Runner(install_type="desktop", result_class=ResultYAML,
                             state_agent=self.state_agent)
        self.partial_state_file_content = partial_state_file_content

    def tearDown(self):
        """
        Clean up the test state file.
        """
        try:
            os.remove(self.state_file)
        except OSError as e:
            if e.errno == 2:  # does not exist
                pass
            else:
                raise

    def test_creation(self):
        self.runner.run()

        self.assertTrue(os.path.exists(self.state_file))

    def test_cleanup(self):
        self.runner.save_state()

        self.assertTrue(os.path.exists(self.state_file),
                        "Runner failed to save state")

        self.state_agent.clean()

        self.assertFalse(os.path.exists(self.state_file),
                         "Failed to remove state file (%s)" % self.state_file)

    def test_status(self):
        """
        Test that a runner properly saves its state during test runs.
        """
        suite_count = self.runner.count_suites()
        test_count = self.runner.count_tests()

        self.runner.run()

        self.assertTrue(os.path.exists(self.state_file))

        self.runner.load_state()

        self.assertEqual(self.runner.count_suites(), suite_count)
        self.assertEqual(self.runner.count_tests(), test_count)

        fp = open(self.state_file, 'r')
        state_data = yaml.load(fp)
        fp.close()

        self.assertEqual(state_data['status'], 'DONE')

    def test_load_state_partial(self):
        """
        Test that a partially run state file can be resumed.
        """
        fp = open(self.state_file, 'w')
        fp.write(self.partial_state_file_content)
        fp.close()

        # Need to set up utah_tests since we're simulating a restart
        test_src = os.path.join(get_module_path(), 'client',
                'examples', 'utah_tests')
        test_dir = os.path.join(UTAH_DIR, 'testsuites',
                                'utah_tests', 'utah_tests')
        if not os.path.exists(test_dir):
            shutil.copytree(test_src, test_dir)

        # Need to set up utah_tests_sample since we're simulating a restart
        test_src = os.path.join(get_module_path(), 'client',
                'examples', 'utah_tests_sample')
        test_dir = os.path.join(UTAH_DIR, 'testsuites', 'utah_tests_sample',
                'utah_tests_sample')
        if not os.path.exists(test_dir):
            shutil.copytree(test_src, test_dir)

        state_agent = StateAgentYAML(state_file=self.state_file)
        runner = Runner(install_type="desktop", state_agent=state_agent,
                        resume=True)
        runner.load_state()

        self.assertEqual(runner.status, 'RUN')

        next_test = runner.get_next_test()

        self.assertEqual(next_test.name, 'test_two')

        runner.run()
