import os
import unittest

from utah.client.exceptions import MissingData
from utah.client.result import ResultYAML
from utah.client.common import (
        UTAH_DIR,
        )
from utah.client import testcase

from .common import (  # NOQA
        setUp,  # Used by nosetests
        tearDown,  # Used by nosetests
        )


class TestTestCase(unittest.TestCase):
    """
    Tests for the TestCase class.
    """
    def setUp(self):
        self.timeout = 10
        self.command = 'echo "command"'
        self.result = ResultYAML()
        self.name = "test_one"
        self.path = "examples/test_one"
        self.suite_dir = os.path.join(UTAH_DIR, 'testsuites', 'examples')

        # Change to the testsuite's working directory
        os.chdir(self.suite_dir)

        self.case = testcase.TestCase(
            name=self.name,
            path=self.path,
            command=self.command,
            timeout=self.timeout,
            result=self.result,
        )

    def test_run(self):
        """
        Test that a TestCase can be run.
        """
        self.case.run()

        # make sure after a testcase run that the cwd hasn't changed.
        self.assertEqual(os.getcwd(), self.suite_dir)
        self.assertTrue(self.case.is_done())

    def test_timeout(self):
        self.assertEqual(self.case.timeout, self.timeout)

    def test_command(self):
        self.assertEqual(self.case.command, self.command)

    def test_save_state(self):
        state = self.case.save_state()

        self.assertEqual(state['status'], 'NOTRUN')
        self.assertEqual(state['command'], self.command)
        self.assertEqual(state['timeout'], self.timeout)

    def test_save_and_load_state(self):
        state = self.case.save_state()

        case = testcase.TestCase(
                name=self.name,
                path=self.path,
                result=ResultYAML()
        )
        case.load_state(state)

        self.assertFalse(case.is_done())

        self.assertEqual(case.name, self.name)
        self.assertEqual(case.path, self.path)
        self.assertEqual(case.command, self.command)
        self.assertEqual(case.timeout, self.timeout)

        case.run()

        self.assertTrue(case.is_done())

    def test_timeout_override(self):
        timeout = 99
        case = testcase.TestCase(
                name=self.name,
                path=self.path,
                timeout=timeout,
                result=ResultYAML()
        )

        self.assertEqual(timeout, case.timeout)

    def test_control_data(self):
        """
        Test that a correctly formed control_data that's passed in
        doesn't raise an exception.
        """
        control_data = {
                'description': 'a test case',
                'command': '/bin/true',
                'dependencies': 'n/a',
                'action': 'n/a',
                'expected_results': 'n/a',
                }

        result = ResultYAML()

        case = testcase.TestCase(
                name=self.name,
                path=self.path,
                result=result,
                _control_data=control_data,
                )

        case.run()
        print result
        result.result()

        self.assertTrue(case.is_done())

    def test_bad_control_data(self):
        """
        Test that an incorrectly formed control_data that's passed in
        does raise an exception.
        """
        control_data = {
                'description': 'a test case',
                'command': '/bin/true',
                }

        result = ResultYAML()

        case = testcase.TestCase(
                name=self.name,
                path=self.path,
                result=result,
                _control_data=control_data,
                )

        with self.assertRaises(MissingData):
            case.run()
