"""
Testsuite specific code.
"""

import jsonschema
import os
import yaml

from .common import run_cmd, parse_control_file
from .common import DEFAULT_TSLIST, DEFAULT_TSCONTROL
from .common import CMD_TS_BUILD, CMD_TS_SETUP, CMD_TS_CLEANUP
from .common import do_nothing
from .testcase import TestCase
from .exceptions import MissingFile, ValidationError


def parse_runlist_file(runlist_file):
    """
    Parse a tslist.run runlist file and check agains schema
    """
    with open(runlist_file, 'r') as fp:
        data = yaml.load(fp)

    jsonschema.validate(data, TestSuite.RUNLIST_SCHEMA)

    return data


class TestSuite(object):
    """
    The TestSuite class.
    """

    build_cmd = None
    timeout = None
    ts_setup = None
    ts_cleanup = None
    control_file = None
    # status is one of:
    # 'NOTRUN', 'BUILD', 'SETUP', 'INPROGRESS', 'CLEANUP', and 'DONE'
    status = "NOTRUN"
    save_state_callback = do_nothing

    RUNLIST_SCHEMA = {
        'type': 'array',
        'items': {
            'type': 'object',
            'properties': {
                'test': {
                    'type': 'string',
                    'required': True,
                    },
                'command': {'type': 'string'},
                'run_as': {'type': 'string'},
                'overrides': {'type': 'object'},
                },
            },
        }

    CONTROL_SCHEMA = {
        'type': 'object',
        'properties': {
            'build_cmd': {'type': 'string'},
            'timeout': {'type': 'integer'},
            'ts_setup': {'type': 'string'},
            'ts_cleanup': {'type': 'string'},
             }
        }

    def __init__(self, name, path, result, runlist_file=DEFAULT_TSLIST,
                 control_file=DEFAULT_TSCONTROL, includes=None, excludes=None,
                 timeout=None,
                 _control_data=None, _runlist_data=None,
                 _save_state_callback=None, _reboot_callback=None):

        self.path = os.path.join(path, name)
        self.reboot_callback = _reboot_callback
        self.passes = 0
        self.failures = 0
        self.errors = 0
        self.timeout = timeout

        # work from within the testsuite's directory.
        # eg. /var/lib/utah/examples
        if not os.path.exists(self.path):
            os.mkdir(self.path)

        os.chdir(self.path)
        """
        Build a TestSuite from a control file's data.
        """
        control_file = os.path.join(name, control_file)

        if os.path.exists(control_file):
            self.control_file = control_file

        self.runlist_file = os.path.join(name, runlist_file)
        self.tests = []
        self.name = name

        if _save_state_callback is not None:
            self.save_state_callback = _save_state_callback

        self.result = result

        control_data = None

        if _control_data is not None:
            control_data = _control_data
        elif self.control_file is not None:
            try:
                control_data = parse_control_file(self.control_file,
                                                  self.CONTROL_SCHEMA)
            except jsonschema.ValidationError as exception:
                raise ValidationError(
                        '{!r} test suite control file invalid: {!r}\n'
                        'Detailed information: {}'
                        .format(self.name, self.control_file, exception))

        if control_data is not None:
            # already initialized
            if self.timeout is not None:
                control_data['timeout'] = self.timeout

            self.__dict__.update(control_data)

        if _runlist_data is not None:
            runlist_data = _runlist_data
        elif os.path.exists(self.runlist_file):
            try:
                runlist_data = parse_runlist_file(self.runlist_file)
            except jsonschema.ValidationError as exception:
                raise ValidationError(
                        '{!r} test suite runlist invalid: {!r}\n'
                        'Detailed information: {}'
                        .format(self.name, self.runlist_file, exception))
        else:
            raise MissingFile('File not found: {}'.format(self.runlist_file))

        for test in runlist_data:
            name = test['test']
            test_path = os.path.join(self.name, name)

            if includes is not None and name not in includes:
                continue

            if excludes is not None and name in excludes:
                continue

            command = test.get('command')

            tc = TestCase(name=name, path=test_path,
                          command=command, timeout=self.timeout,
                          result=self.result,
                          _save_state_callback=self.save_state_callback,
                          _reboot_callback=self.reboot_callback)
            if 'overrides' in test:
                if self.timeout is not None:
                    test['overrides']['timeout'] = self.timeout

                tc.process_overrides(test['overrides'])

            self.tests.append(tc)

    def __str__(self):
        return "%s: %s" % (self.control_file, self.runlist_file)

    def set_status(self, status):
        """
        Set the status for the test suite and call the save state callback.
        """
        self.status = status
        self.save_state_callback()

    def build(self, result):
        """
        Run build, but only if we haven't started a run yet.
        """
        if self.status == 'NOTRUN':
            self.set_status('BUILD')
            result.add_result(run_cmd(self.build_cmd, cwd=self.name,
                                      cmd_type=CMD_TS_BUILD))

    def setup(self, result):
        """
        Run ts_setup, but only if build() has just passed.
        """
        if self.status == 'BUILD' and result.status == 'PASS':
            self.set_status('SETUP')
            result.add_result(run_cmd(self.ts_setup, cwd=self.name,
                                      cmd_type=CMD_TS_SETUP))

    def cleanup(self, result):
        """
        Run ts_cleanup after a run.
        """
        if self.status == 'INPROGRESS':
            self.set_status('CLEANUP')
            result.add_result(run_cmd(self.ts_cleanup, cwd=self.name,
                                      cmd_type=CMD_TS_CLEANUP))

    def run(self):
        """
        Run the test cases; including any build, setup, and cleanup commands.
        """

        # Return value to indicate whether processing of a Runner should
        # continue.  This is to avoid a shutdown race on reboot cases.
        keep_going = True

        # Work from the testsuite directory
        os.chdir(self.path)

        result = self.result

        # keep track of which testsuite we are running tests in
        self.result.testsuite = self.path

        self.build(result)
        self.setup(result)

        # Always enter this loop since test.run() will handle checking if
        # the test has already been run.
        self.status = "INPROGRESS"
        if result.status == 'PASS':
            for test in self.tests:
                keep_going = test.run()

                if test.run_status == 'PASS':
                    self.passes += 1
                elif test.run_status == 'FAIL':
                    self.failures += 1
                elif test.run_status == 'ERROR':
                    self.errors += 1

                if not keep_going:
                    return keep_going

        self.cleanup(result)

        if self.status == 'CLEANUP':
            self.set_status("DONE")

        return keep_going

    def add_test(self, tests):
        """
        Add a single test or list of tests to this suite.
        """
        if isinstance(tests, TestCase):
            self.tests.append(tests)
        else:
            try:
                for test in tests:
                    self.add_test(test)
            except TypeError:
                pass

    def count_tests(self):
        return len(self.tests)

    def load_state(self, state):
        """
        Restore our state from the supplied dictionary.

        Requires that the fieldnames in the dictionary match the class
        properties.
        """
        # XXX: Should this be done explicitly?
        self.__dict__.update(state)

        self.tests = []
        for state_test in state['tests']:
            test = TestCase(name=state_test['name'], path=state_test['path'],
                            result=self.result,
                            _save_state_callback=self.save_state_callback,
                            _reboot_callback=self.reboot_callback)
            test.load_state(state_test)
            self.tests.append(test)

    def save_state(self):
        state = {
            'name': self.name,
            'status': self.status,
            'passes': self.passes,
            'failures': self.failures,
            'errors': self.errors,
            'tests': [],
            'ts_setup': self.ts_setup,
            'ts_cleanup': self.ts_cleanup,
            'build_cmd': self.build_cmd,
            'timeout': self.timeout,
        }

        for test in self.tests:
            state['tests'].append(test.save_state())

        return state

    def is_done(self):
        """
        Determine if the suite is done.  This might mean that something has
        failed.  Used by Runner to determine if the suite needs to be re-run
        on resume.
        """
        return self.status == 'DONE' or self.status == 'CLEANUP'

    def get_next_test(self):
        """
        Return the next test to be run.

        Mainly used for debugging.
        """

        test = None

        for t in self.tests:
            if not t.is_done():
                test = t
                break

        return test
