#!/usr/bin/python
"""
Provide config variables via utah.config.
When run directly as part of the package build, output machine-independent
defaults as json for writing a config file.
"""

import getpass
import json
import logging
import os
import socket
import subprocess


def getbridge():
    """
    Get virbr0 if it is set in the qemu config or the utah config.
    Use this (or the default virbr0) as the bridge for bridged network VMs.
    """
    dns = os.path.join('/', 'etc', 'utah', 'dns')
    if os.path.isfile(dns):
        return (subprocess.Popen([dns], stdout=subprocess.PIPE)
                .communicate()[0].strip())
    else:
        return 'virbr0'

DEFAULTS = dict(
    arch='i386',
    boottimeout=30,
    checktimeout=180,
    consoleloglevel=logging.WARNING,
    debug=False,
    fileloglevel=logging.INFO,
    group='utah',
    image=None,
    initrd=None,
    installtimeout=3600,
    installtype='desktop',
    kernel=None,
    logpath=os.path.join('/', 'var', 'log', 'utah'),
    machineid=None,
    name=None,
    new=False,
    nfscommand=['sudo', os.path.join('/', 'etc', 'init.d', 'nfs-kernel-server'), 'reload'],
    nfsconfigfile=os.path.join('/', 'etc', 'exports'),
    nfsoptions='*(ro,async,no_root_squash,no_subtree_check)',
    powertimeout=5,
    preseed=None,
    qemupath='qemu:///system',
    series='precise',
    template=None,
)

"""
These depend on the local user/path, and need to be filtered out of the config
file created at package build time.
"""
LOCALDEFAULTS = dict(
    bridge=getbridge(),
    hostname=socket.gethostname(),
    sshprivatekey=os.path.join(os.path.expanduser('~'), '.ssh', 'utah'),
    sshpublickey=os.path.join(os.path.expanduser('~'), '.ssh', 'utah.pub'),
    sshknownhosts=os.path.join(os.path.expanduser('~'), '.ssh', 'known_hosts'),
    user=getpass.getuser(),
    vmpath=os.path.join(os.path.expanduser('~'), 'vm'),
)

"""
These depend on other config options, so they're added last.
"""
LOCALDEFAULTS['logfile'] = os.path.join(DEFAULTS['logpath'],
                                        LOCALDEFAULTS['hostname'] + '.log')
LOCALDEFAULTS['debuglog'] = os.path.join(DEFAULTS['logpath'],
                                    LOCALDEFAULTS['hostname'] + '-debug.log')

DEFAULTS.update(LOCALDEFAULTS)

CONFIG = {}
CONFIG.update(DEFAULTS)

"""
Process config files.
"""
files = ['/etc/utah/config', '~/.utah/config', os.getenv('UTAH_CONFIG_FILE')]
for conffile in files:
    if conffile is not None:
        myfile = os.path.expanduser(conffile)
        if os.path.isfile(myfile):
            CONFIG.update(json.load(open(myfile)))

locals().update(CONFIG)


def dumpdefaults(build=False):
    if build:
        """
        Remove user and path dependent items so they won't get set on the
        machine that builds the package.
        """
        for item in LOCALDEFAULTS:
            DEFAULTS.pop(item)
    return json.dumps(DEFAULTS, sort_keys=True, indent=4)


def dumpconfig():
    return json.dumps(CONFIG, sort_keys=True, indent=4)


if __name__ == '__main__':
    """
    If we're not told otherwise, assume this is the package build.
    """
    print(dumpdefaults(build=True))
