"""
All commands use bsdtar and accept a logmethod to use for logging.
"""
#TODO: Make an iso class that can return things like arch, series, type, etc.
#TODO: Add the existing functions to the iso class.
#TODO: Convert the code to use the class instead of the existing functions.
import logging
import logging.handlers
import os
import subprocess
import sys
import urllib

from utah import config
from utah.exceptions import UTAHException


def listfiles(image, logmethod=None, returnlist=False):
    """
    Return either a subprocess instance listing the contents of an iso, or a
    list of files in the iso if returnlist is True.
    """
    cmd = ['bsdtar', '-t', '-f', image]
    if logmethod is not None:
        logmethod('bsdtar list command: ' + ' '.join(cmd))
    if returnlist:
        return subprocess.check_output(cmd).strip().split('\n')
    else:
        return subprocess.Popen(cmd, stdout=subprocess.PIPE,
                                stderr=subprocess.PIPE).communicate()


def getrealfile(image, filename, outdir=None, logmethod=None):
    """
    Return a command to safely extract a file from an iso.
    Based on unbsdtar-safelink from ubuntu iso testing.
    """
    cmd = ['bsdtar', '-t', '-v', '-f', image, filename]
    if logmethod is not None:
        logmethod('bsdtar list command: ' + ' '.join(cmd))
    output = (subprocess.Popen(cmd, stdout=subprocess.PIPE)
              .communicate()[0].split('\n')[0].split())
    try:
        if output[0].startswith('l'):
            if outdir is None:
                cmd = ['ln', '-s', output[-1], output[-3]]
            else:
                cmd = ['ln', '-s', output[-1],
                       os.path.join(outdir, output[-3])]
        else:
            realfile = output[-1]
            if outdir is None:
                cmd = ['bsdtar', '-x', '-f', image, '-O', realfile]
            else:
                cmd = ['bsdtar', '-x', '-f', image, '-C', outdir,
                       '-O', realfile]
    except IndexError:
        raise UTAHException('Cannot unpack image: ' + image)
    if logmethod is not None:
        logmethod('bsdtar extract command: ' + ' '.join(cmd))
    return cmd


def extract(image, filename, outdir=None, outfile=None, logmethod=None, **kw):
    """
    Return a subprocess to extract a file from an iso, passing any additional
    keyword arguments to the subprocess.
    If outfile is None, extract to standard out.
    If outdir is present and outfile is None, extract the file to outdir with
    the same name.
    If outdir and outfile are both present, extract to outfile in outdir.
    If the file is a hardlink, extract the actual file.
    If the file is a symlink, return a command to create the symlink.
    """
    if outfile is None:
        if outdir is None:
            cmd = getrealfile(image, filename, logmethod=logmethod)
            # The original function was designed to send the file over stdout,
            # this removes that
            # I think the updated function takes care of this already
            # if '-C' in cmd:
            #     cmd.remove('-C')
            if cmd[-1] != filename:
                if logmethod is not None:
                    logmethod(filename + ' appears to be a link to ' + cmd[-1])
                dirname = os.path.dirname(filename)
                if not os.path.isdir(dirname):
                    os.makedirs(dirname)
                os.chmod(dirname, 0775)
                if os.path.isfile(filename):
                    os.chmod(filename, 0775)
                return subprocess.call(cmd, stdout=open(filename, 'w'), **kw)
        else:
            cmd = getrealfile(image, filename,
                              outdir=outdir, logmethod=logmethod)
        if '-O' in cmd:
            cmd.remove('-O')
        return subprocess.Popen(cmd, stdout=subprocess.PIPE,
                                stderr=subprocess.PIPE, **kw).communicate()
    else:
        cmd = getrealfile(image, filename, logmethod=logmethod)
        return subprocess.call(cmd, stdout=open(outfile, 'w'), **kw)


def dump(image, filename, logmethod=None, **kw):
    """
    Return a subprocess dumping a file in an iso to standard out, passing any
    additional keyword arguments to the subprocess.
    """
    cmd = getrealfile(image, filename, logmethod=logmethod)
    return subprocess.Popen(cmd, stdout=subprocess.PIPE, **kw)


class ISO(object):
    """
    Provide a simplified method of interfacing with images.
    """
    def __init__(self, image, dlpercentincrement=1, logger=None):
        self.dlpercentincrement = dlpercentincrement
        if logger is None:
            self._loggersetup()
        else:
            self.logger = logger
        if image.startswith('~'):
            path = os.path.expanduser(image)
            self.logger.debug('Rewriting ~-based path: ' + image +
                              ' to: ' + path)
        else:
            path = image

        self.percent = 0
        self.logger.info('Preparing image: ' + path)
        self.image = urllib.urlretrieve(path, reporthook=self.dldisplay)[0]
        self.logger.debug(path + ' is locally available as ' + self.image)
        self.installtype = self.getinstalltype()
        if self.installtype == 'mini':
            self.logger.debug('Using mini image')
            self.infofile = '.disk/mini-info'
        else:
            self.logger.debug('Using normal image')
            self.infofile = './.disk/info'
        # Info file looks like this:
        # Ubuntu 12.04 LTS "Precise Pangolin" - Release i386 (20120423)
        # or this:
        # Ubuntu 12.10 "Quantal Quetzal" - Alpha amd64 (20120820)
        # i.e. Ubuntu Version (LTS) "Series Codename" - Alpha/Beta/Release
        # arch (buildnumber)
        self.arch = self.getarch()
        self.series = self.getseries()

    def _loggersetup(self):
        """
        Initialize the logging for the image.
        """
        self.logger = logging.getLogger('utah.iso')
        self.logger.propagate = False
        self.logger.setLevel(logging.INFO)
        for handler in list(self.logger.handlers):
            self.logger.removeHandler(handler)
            del handler
        self.consolehandler = logging.StreamHandler(stream=sys.stderr)
        console_formatter = logging.Formatter('%(levelname)s: %(message)s')
        self.consolehandler.setFormatter(console_formatter)
        self.consolehandler.setLevel(config.consoleloglevel)
        self.logger.addHandler(self.consolehandler)
        self.filehandler = logging.handlers.WatchedFileHandler(config.logfile)
        file_formatter = logging.Formatter('%(asctime)s iso '
                                           + ' %(levelname)s: %(message)s')
        self.filehandler.setFormatter(file_formatter)
        self.filehandler.setLevel(config.fileloglevel)
        self.logger.addHandler(self.filehandler)
        if config.debuglog is not None:
            self.logger.setLevel(logging.DEBUG)
            self.debughandler = (
                    logging.handlers.WatchedFileHandler(config.debuglog))
            self.debughandler.setFormatter(file_formatter)
            self.debughandler.setLevel(logging.DEBUG)
            self.logger.addHandler(self.debughandler)

    def getinstalltype(self):
        """
        Inspect the image's files to get the image type.
        If .disk/mini-info exists, it's mini.
        If the casper directory exists, it's desktop.
        If ubuntu-server.seed exists in the preseeds directory, it's server.
        """
        self.logger.info('Getting image type of ' + self.image)
        files = set(self.listfiles(returnlist=True))
        installtype = 'alternate'
        if '.disk/mini-info' in files:
            installtype = 'mini'
        elif 'casper' in files or './casper' in files:
            installtype = 'desktop'
        elif './preseed/ubuntu-server.seed' in files:
            installtype = 'server'
        self.logger.info('Image type is: ' + installtype)
        return installtype

    def getarch(self):
        """
        Unpack the image's info file to get the arch.
        """
        stdout = self.dump(self.infofile).communicate()[0]
        arch = stdout.split()[-2]
        self.logger.info('Arch is: ' + arch)
        return arch

    def getseries(self):
        """
        Unpack the image's info file to get the series.
        """
        stdout = self.dump(self.infofile).communicate()[0]
        for word in stdout.split():
            if word.startswith('"'):
                series = word.strip('"').lower()
                break
        self.logger.info('Series is ' + series)
        return series

    def dldisplay(self, blocks, size, total):
        read = blocks * size
        percent = 100 * read / total
        if percent >= self.percent:
            self.logger.info('File ' + str(percent) + '% downloaded')
            self.percent += self.dlpercentincrement
        self.logger.debug("%s read, %s%% of %s total" % (read, percent, total))

    def listfiles(self, returnlist=False):
        return listfiles(self.image, self.logger.debug, returnlist=returnlist)

    def getrealfile(self, filename, outdir=None):
        return getrealfile(self.image, filename, outdir=outdir,
                           logmethod=self.logger.debug)

    def extract(self, filename, outdir=None, outfile=None, **kw):
        return extract(self.image, filename, outdir, outfile,
                       logmethod=self.logger.debug, **kw)

    def dump(self, filename, **kw):
        return dump(self.image, filename, logmethod=self.logger.debug, **kw)

    def extractall(self):
        for myfile in self.listfiles(returnlist=True):
            self.extract(myfile)
