#!/usr/bin/python

#
# Copyright (C) 2012, Canonical Ltd (http://www.canonical.com/)
#
# This file is part of utah.
#
# iso_satatic_validation is free software: you can redistribute it
# and/or modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation, either version 3 of
# the License, or (at your option) any later version.
#
# iso_satatic_validation is distributed in the hope that it will
# be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
# of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with ubuntu-server-iso-testing.  If not, see
# <http://www.gnu.org/licenses/>.
#

import os
import argparse
import sys
import logging
import urllib2
import subprocess
import hashlib
import unittest
import re
import tempfile
import shutil

lib_path = os.path.abspath('../')
sys.path.append(lib_path)
import iso


# Defaults
DEFAULT_ISOROOT = os.path.expanduser('~/Downloads')
DEFAULT_URL = 'http://cdimage.ubuntu.com/'

#constants
ONE_MB_BLOCK = 2 ** 20

#Read the iso name and initialise global variables
parser = argparse.ArgumentParser()
parser.add_argument('--name')
args = parser.parse_args()

logging.basicConfig(level=logging.INFO)

if args.name is None or '.iso' not in args.name:
    print 'The usage:', sys.argv[0],  '--name release-variant-arch.iso'
    sys.exit(1)
else:
    st_name = args.name.rstrip('.iso')
    try:
        (st_release, st_variant, st_arch) = st_name.rsplit('-')
    except ValueError:
        sys.exit("Image name be in the form of release-variant-arch.iso")
    logging.debug("The file being tested is: '%s'", args.name)

#Set a flag for ubiquity based images
if st_variant == 'desktop' or st_variant == 'dvd':
    ubiquity_image = True
else:
    ubiquity_image = False


class TestValidateISO(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        try:
            cls.temp_dir = tempfile.mkdtemp()
        except OSError:
            logging.error("Creating temp dirrectory failed")
            unittest.fail("Setup error")

    def setUp(self):
        self.iso_name = args.name
        self.block_size = ONE_MB_BLOCK
        self.iso_location = os.path.join(DEFAULT_ISOROOT, args.name)
        self.st_release = st_release
        self.st_variant = st_variant
        self.st_arch = st_arch

        if self.st_release != 'precise':
            self.url = DEFAULT_URL
        else:
            self.url = os.path.join(DEFAULT_URL, 'precise')

        if self.st_variant == 'desktop':
            self.url = os.path.join(self.url, 'daily-live')
        elif self.st_variant == 'dvd':
            self.url = os.path.join(self.url, 'dvd')
        elif self.st_variant == 'alternate':
            self.url = os.path.join(self.url, 'daily')
        elif self.st_variant == 'server':
            if self.st_release == 'precise':
                self.url = os.path.join(DEFAULT_URL, 'ubuntu-server',
                                        'precise', 'daily')
            else:  # current dev release
                self.url = os.path.join(DEFAULT_URL, 'ubuntu-server', 'daily')
        else:
            self.fail("Image name be in the form of release-variant-arch.iso")

    # Test if the sha check sum of the iso matches that is given in the server
    def test_sha256_checksum(self):
        current_iso_digest = {}
        sha256sum_url = os.path.join(self.url, 'current', 'SHA256SUMS')
        try:
            fh = urllib2.urlopen(sha256sum_url)
        except urllib2.HTTPError, e:
            logging.error("Failed to fetch URL '%s': %s. Aborting!",
                          sha256sum_url, e)
            self.fail("Failed to fetch sha checksum from the server")

        # Process the file (contains a number of entries containing iso
        # names and the corresponding checksums)
        for digest in fh:
            (chksum, sep,  iso_image) = digest.partition(' *')
            current_iso_digest[iso_image.rstrip()] = chksum
        fh.close()

        try:
            iso_sha256 = current_iso_digest[self.iso_name]
        except KeyError:
            logging.error("No digest found for image '%s'", self.iso_name)
            self.fail("The iso does not have a matching digest")

        # Validate checksum of file fname against sha256 hash
        logging.debug("Calculating hash for file '%s'", self.iso_name)
        sha256 = hashlib.sha256()
        with open(self.iso_location) as f:
            while True:
                data = f.read(self.block_size)
                if not data:
                    break
                sha256.update(data)
        sha256 = sha256.hexdigest()
        logging.debug("Expected Checksum: '%s'", iso_sha256)
        logging.debug("Local File Checksum: '%s'", sha256)
        self.assertEqual(iso_sha256, sha256)

    # Test if the list file in the repository matches the content of the iso
    def test_files_list(self):
        if self.st_variant == 'server':
            list_url = os.path.join(self.url, 'current',
                                    self.st_release + '-' + 'server' + '-' +
                                    self.st_arch + '.list')
        else:
            list_url = os.path.join(self.url, 'current',
                                    self.st_release + '-' +
                                    self.st_variant + '-' +
                                    self.st_arch + '.list')

        try:
            list_repository = urllib2.urlopen(list_url)
        except urllib2.HTTPError, e:
            logging.error("Failed to fetch URL '%s': %s . Aborting!",
                          list_url, e)
            self.fail("Failed to fetch files list from the server")

        logging.debug('Extracting list file info from ISO')
        (stdout, stderr) = iso.listfiles(self.iso_location)
        logging.debug('Checking for error in extracting the file list')
        self.assertEqual(stderr, '')

        for list_entry in list_repository:
            logging.debug('Checking that each list entry is in the iso')
            self.assertIn(list_entry[1:].rstrip(), stdout)

    # Test if the manfest is the same as that is given in the server
    @unittest.skipUnless(ubiquity_image,
                        "manifest only for ubiquity based images")
    def test_manifest(self):
        manifest_url = os.path.join(self.url, 'current',
                                    self.st_release + '-' +
                                    self.st_variant + '-' +
                                    self.st_arch + '.manifest')
        try:
            manifest = urllib2.urlopen(manifest_url)
        except urllib2.HTTPError, e:
            logging.error("Failed to fetch URL '%s': %s . Aborting!",
                          manifest_url, e)
            self.fail("Failed to fetch manifest file from the server")

        logging.debug('Extracting manifest from the iso')
        (stdout, stderr) = iso.extract(self.iso_location,
                                       'casper/filesystem.manifest',
                                       self.temp_dir)
        logging.debug('Check for error in extracting manifest info from iso')
        self.assertEqual(stderr, '')

        manifest_path = os.path.join(self.temp_dir,
                                     'casper/filesystem.manifest')
        manifest_local = open(manifest_path)

        for manifest_entry in manifest:
            logging.debug('Checking manifest are the same in iso and repo')
            self.assertIn(manifest_entry, manifest_local)

    # Test if the buildid present in the download server is
    # the same as that given in distro info
    def test_build_id(self):
        logging.debug('Extract distro information of the image from the iso')
        (stdout, stderr) = iso.extract(self.iso_location, "./.disk/info",
                                       self.temp_dir)
        logging.debug('Checking for error in extracting distro info from iso')
        self.assertEqual(stderr, '')

        buildid_path = os.path.join(self.temp_dir, './.disk/info')
        buildinfo = open(buildid_path)

        try:
            buildid_page = urllib2.urlopen(self.url).read()
        except urllib2.HTTPError, e:
            logging.error("Failed to fetch URL '%s': %s. Abort!", self.url, e)
            self.fail("Failed to fetch build page info from the server")

        logging.debug('Look for the current buildid from the downloaded page')
        buildid_page_lines = re.findall('href=[\'"]?([^\'" >]+)',
                                        buildid_page)
        for ids_server in buildid_page_lines:
            if ids_server == 'current/':
                break
            id_server = ids_server.rstrip('/')

        for build_info_line in buildinfo:
            logging.debug('Checking buildids are the same in iso and repo')
            self.assertIn(id_server, build_info_line)

    # Test if  wubi is present for desktop or dvd and it is a valid PE
    @unittest.skipUnless(st_variant == 'desktop' and
                         st_arch != 'powerpc', "Skipping non ubiquity images")
    def test_wubi(self):
        logging.debug('Extracting wubi from the desktop images')
        (stdout, stderr) = iso.extract(self.iso_location, './wubi.exe',
                                       self.temp_dir)
        logging.debug('Check for error in extracting wubi.exe from the iso')
        self.assertEqual(stderr, '')

        wubi_path = os.path.join(self.temp_dir, './wubi.exe')

        cmd = ["file", wubi_path]
        output = subprocess.Popen(cmd, stdout=subprocess.PIPE,
                                  stderr=subprocess.PIPE)
        (stdout, stderr) = output.communicate()
        logging.debug('Check that wubi has a valid PE header')
        self.assertEqual(stderr, '')
        self.assertIn(" PE32 executable (GUI)", stdout)

    # Test if the relevant files are present in the iso for desktop iso
    @unittest.skipUnless(st_variant == 'desktop' and
                         st_arch != 'powerpc', "Skipping for non desktop iso")
    def test_files_ubiquity(self):
        (stdout, stderr) = iso.listfiles(self.iso_location,
                                         logmethod=logging.debug)
        logging.debug('Check for error in extracting file list from the iso')
        self.assertEqual(stderr, '')
        files_list = open("data/file_list_ubiquity")
        for list_server in files_list:
            logging.debug('Check if relevant files are present in the iso')
            self.assertIn(list_server.rstrip(), stdout)

    # Test if the relevant important files are present in the iso for dvd iso
    @unittest.skipUnless(st_variant == 'dvd',
                        "This file list test is only specific for dvd")
    def test_files_ubiquity_dvd(self):
        (stdout, stderr) = iso.listfiles(self.iso_location,
                                         logmethod=logging.debug)
        logging.debug('Check for error in extracting file list from the iso')
        self.assertEqual(stderr, '')
        files_list = open("data/file_list_ubiquity_dvd")
        for list_entry in files_list:
            logging.debug('Check if relevant files are present in the iso')
            self.assertIn(list_entry.rstrip(), stdout)

    # Test if the relevant files are present in the iso for d-i based images,
    # two list of files are used
    # file_list_di for i386 and amd64
    # file_list_di_powerpc for powerpc and amd64+mac
    # Test is skipped for ubiquity based images
    @unittest.skipIf(ubiquity_image, "Skipping for ubiquity images")
    def test_files_di(self):
        (stdout, stderr) = iso.listfiles(self.iso_location,
                                         logmethod=logging.debug)
        logging.debug('Check for error in extracting file list from the iso')
        self.assertEqual(stderr, '')

        if self.st_arch == 'powerpc' or self.st_arch == 'amd64+mac':
            files_list = open("data/file_list_di_powerpc")
        else:  # i386 or amd64
            files_list = open("data/file_list_di")

        for list_server in files_list:
            logging.debug('check if important d-i files are present in iso')
            self.assertIn(list_server.rstrip(), stdout)

    # Test if vmlinuz is a valid executable
    @unittest.skipUnless(ubiquity_image and st_arch != 'powerpc',
                         "vmlinuz is present only for ubiquity based images")
    def test_vmlinuz(self):
        (stdout, stderr) = iso.extract(self.iso_location, 'casper/vmlinuz',
                                       self.temp_dir)
        logging.debug('Asserting that vmlinuz is present in the iso')
        self.assertEqual(stderr, '')

        vmlinuz_path = os.path.join(self.temp_dir, 'casper/vmlinuz')

        cmd = ["file", vmlinuz_path]
        output = subprocess.Popen(cmd, stdout=subprocess.PIPE,
                                  stderr=subprocess.PIPE)
        (stdout, stderr) = output.communicate()
        logging.debug('Check if vmlinuz present in the iso is a Linux kernel')
        self.assertEqual(stderr, '')
        self.assertIn("Linux kernel", stdout)

    # Test if filesystem.squashfs is present and
    # if that is a squash fs for ubiquity based images
    @unittest.skipUnless(ubiquity_image, "Skipping for non ubiquity images")
    def test_filesystem_squashfs(self):
        logging.debug('Extracting the filesystem.squashfs')
        (stdout, stderr) = iso.extract(self.iso_location,
                                       'casper/filesystem.squashfs',
                                       self.temp_dir)
        logging.debug('Check for error in extracting filesystem.squashfs')
        self.assertEqual(stderr, '')

        squashfs_path = os.path.join(self.temp_dir,
                                     'casper/filesystem.squashfs')

        logging.debug('Checking the file is an actual .squashfs')
        cmd = ["file", squashfs_path]
        output = subprocess.Popen(cmd, stdout=subprocess.PIPE,
                                  stderr=subprocess.PIPE)
        (stdout, stderr) = output.communicate()
        logging.debug('Check if signature contains \"Squashfs filesystem\"')
        self.assertEqual(stderr, '')
        self.assertIn("Squashfs filesystem", stdout)

        #unsquashfs extracts only in pwd
        os.chdir(self.temp_dir)

        cmd = ["unsquashfs", "-f",  squashfs_path, "-l", "vmlinuz"]
        output = subprocess.Popen(cmd, stdout=subprocess.PIPE,
                                  stderr=subprocess.PIPE)
        (stdout, stderr) = output.communicate()
        logging.debug('Check that vmlinuz could be extracted')
        self.assertEqual(stderr, '')
        self.assertIn("created 1 directories", stdout)

    # Test if the report.html exists for the d-i based images
    # and that it does not have any conflicting packages
    @unittest.skipIf(ubiquity_image, "Skipping for ubiquity based images")
    def test_report_html(self):
        report_url = os.path.join(self.url, 'current', 'report.html')
        try:
            fh_report_page = urllib2.urlopen(report_url).read()
        except urllib2.HTTPError, e:
            logging.error("Failed to fetch URL '%s': %s .", report_url, e)
            self.fail("Failed to fetch report.html from  the server")

        # check the number of failed packages
        report_page_lines = re.findall('there were ["]?([^" >]+)',
                                       fh_report_page)
        logging.debug('Check for conflicting packages in the report file')
        self.assertEqual(['0'], report_page_lines)

    # clean up
    @classmethod
    def tearDownClass(cls):
        try:
            shutil.rmtree(cls.temp_dir, ignore_errors=False)
        except OSError:
            logging.error("Removing the temp directory failed")
            unittest.fail("Cleanup error")

if __name__ == '__main__':
    from test import test_support
    test_support.run_unittest(TestValidateISO)
