"""
Support bare metal provisioning through cobbler.
"""

import netifaces
import os
import pipes
import shutil
import tempfile
import time

from utah import config
from utah.provisioning.provisioning import (
        CustomInstallMixin,
        Machine,
        SSHMixin,
        )
from utah.provisioning.baremetal.exceptions import UTAHBMProvisioningException
from utah.retry import retry


class CobblerMachine(CustomInstallMixin, SSHMixin, Machine):
    """
    Provide a class to provision a machine via cobbler.
    """
    # TODO: may need to fix DHCP/hostname issues, may just be magners
    def __init__(self, boot=None, cargs=None, inventory=None,
                 name=None, preseed=None, *args, **kw):
        if name is None:
            raise UTAHBMProvisioningException(
                    'Machine name reqired for cobbler machine')
        if cargs is None:
            raise UTAHBMProvisioningException(
                    'No cobbler arguments given for machine creation')
        else:
            self.cargs = cargs
        self.inventory = inventory
        if preseed is None:
            preseed = '/etc/utah/default-preseed.cfg'
        super(CobblerMachine, self).__init__(*args, name=name,
                                             preseed=preseed, **kw)
        if self.image is None:
            raise UTAHBMProvisioningException(
                    'Image file required for cobbler installation')
        self._custominit(arch=self.arch, boot=boot,
                         installtype=self.installtype, series=self.series)
        # TODO: verify we have nfs support for desktop image

        # TODO: Rework cinitrd to be less of a confusing collection of kludges
        self.cinitrd = None
        if self.installtype in ['alternate', 'server']:
            self.cinitrd = os.path.join('install', 'netboot',
                                        'ubuntu-installer', self.arch,
                                        'initrd.gz')
        if self.arch == 'amd64':
            self.carch = 'x86_64'
        else:
            self.carch = self.arch
        self.cname = self.name + '-' + self.carch
        self.logger.debug('Cobbler arch is ' + self.carch)
        self.logger.debug('Cobbler machine init finished')

    def _provision(self, checktimeout=config.checktimeout,
                   installtimeout=config.installtimeout):
        """
        Install a machine.
        """
        # TODO: support for reusing existing machines
        self.tmpdir = tempfile.mkdtemp(prefix='/tmp/' + self.name + '_')
        os.chdir(self.tmpdir)

        # TODO: try to remove this step, mount the iso and import that
        # with a specified kernel/preseed/initrd from the tmpdir
        if self.installtype in ['alternate', 'desktop', 'server']:
            self.logger.info('Extracting image to ' + self.tmpdir)
            self.image.extractall()

        kernel = self._preparekernel()
        if self.cinitrd is None:
            cinitrd = None
        else:
            cinitrd = os.path.join(self.tmpdir, self.cinitrd)
        initrd = self._prepareinitrd(initrd=cinitrd)
        self._unpackinitrd(initrd=initrd)
        self._setuplatecommand()
        if self.installtype == 'desktop':
            self.logger.info('Configuring latecommand for desktop')
            myfile = open(os.path.join(self.tmpdir, 'initrd.d',
                          'utah-latecommand'), 'a')
            myfile.write(
"""
chroot /target sh -c 'sed "/eth[0-9]/d" -i /etc/network/interfaces'
cp /etc/resolv.conf /target/etc/resolv.conf
""")
            myfile.close()
        self._setuppreseed()
        initrd = self._repackinitrd()

        if cinitrd is not None and initrd != cinitrd:
            os.chmod(cinitrd, 0755)
            shutil.copyfile(initrd, cinitrd)

        self.logger.info('Setting up system with cobbler')
        self.logger.debug('Removing old system')
        self._cobble(['system', 'remove', '--name=' + self.name])
        self.logger.info('Removing old profile')
        self._cobble(['profile', 'remove', '--name=' + self.cname])
        self.logger.info('Removing old distro')
        self._cobble(['distro', 'remove', '--name=' + self.cname])

        preseed = os.path.join(self.tmpdir, 'initrd.d', 'preseed.cfg')

        if self.installtype in ['alternate', 'server']:
            # TODO: support more image types,
            # maybe do this without unpacking ISO
            self.logger.info('Importing image')
            self._cobble(['import', '--name=' + self.cname,
                          '--path=' + self.tmpdir, '--arch=' + self.carch])
        elif self.installtype in ['mini', 'desktop']:
            self.logger.info('Creating distro')
            self._cobble(['distro', 'add', '--name=' + self.cname,
                          '--kernel=' + kernel, '--initrd=' + initrd])
            self.logger.info('Creating profile')
            self._cobble(['profile', 'add', '--name=' + self.cname,
                          '--distro=' + self.cname])

        if self.installtype == 'desktop':
            self.logger.info('Setting up NFS for desktop install')
            self.logger.debug('Adding export to NFS config file')
#            nfsfile = open(config.nfsconfigfile, 'a')
#            nfsfile.write("\n" + self.tmpdir + ' ' + config.nfsoptions + "\n")
#            nfsfile.close()
            pipe = pipes.Template()
            pipe.append('sudo tee -a ' + config.nfsconfigfile + ' >/dev/null',
                        '-.')
            pipe.open('/dev/null', 'w').write(self.tmpdir + ' ' +
                      config.nfsoptions + "\n")
            self.logger.debug('Reloading NFS config')
            self._runargs(config.nfscommand)
            self.logger.debug('Adding NFS boot options')
            try:
                iface = config.nfsiface
            except AttributeError:
                iface = config.bridge
            ip = netifaces.ifaddresses(iface)[netifaces.AF_INET][0]['addr']
            self.cmdline += (' root=/dev/nfs netboot=nfs nfsroot=' +
                             ip + ':' + self.tmpdir)
            self.cmdline = self.cmdline.strip()

        self.logger.info('Adding kernel boot options to distro')
        self._cobble(['distro', 'edit', '--name=' + self.cname,
                      '--kopts=' + self.cmdline])

        self.logger.info('Adding kickstart to profile')
        self._cobble(['profile', 'edit', '--name=' + self.cname,
                      '--kickstart=' + preseed])

        self.logger.info('Adding system')
        self._cobble(['system', 'add', '--name=' + self.name,
                      '--profile=' + self.cname, '--netboot-enabled=Y']
                      + ['--' + arg + '=' + self.cargs[arg]
                         for arg in self.cargs])

        # Things seem to be working fine without this, but sometimes things still get broken
        #self.logger.info('Syncing cobbler')
        #self._cobble(['sync'])
        try:
            self._runargs(config.cobblerfix)
        except AttributeError:
            pass
        self.restart()

        self.logger.info('Waiting for installation to begin')
        time.sleep(config.boottimeout)
        self._cobble(['system', 'edit', '--name=' + self.name,
                      '--netboot-enabled=N'])

        self.logger.info('System installing')
        self.logger.info('Checking every ' + str(checktimeout) +
                         ' seconds until system is installed')
        self.logger.info('Will time out after ' + str(installtimeout)
                         + ' seconds')
        retry(self.sshcheck, checktimeout, logmethod=self.logger.info)

        if self.debug:
            self.logger.info('Leaving temp directory '
                             'because debug is enabled: ' + self.tmpdir)
        else:
            self.logger.info('Cleaning up temp directory')
            if self.installtype == 'desktop':
                self.logger.info('Removing NFS share')
                self._removenfs()
            # Cribbed from http://svn.python.org
            # /projects/python/trunk/Mac/BuildScript/build-installer.py
            for dirpath, dirnames, filenames in os.walk(self.tmpdir):
                for name in (dirnames + filenames):
                    filename = os.path.join(dirpath, name)
                    if not os.path.islink(filename):
                        self.logger.debug('Changing permissions of '
                                          + filename)
                        os.chmod(filename, 0775)
            shutil.rmtree(self.tmpdir)

        self.provisioned = True
        self.active = True
        uuid_check_command = ('[ "{uuid}" == "$(cat /etc/utah/uuid)" ]'
                              .format(uuid=self.uuid))
        if self.run(uuid_check_command)[0] != 0:
            self.provisioned = False
            raise UTAHBMProvisioningException(
                    'Installed UUID differs from CobblerMachine UUID; '
                    'installation probably failed. '
                    'Try restarting cobbler or running cobbler sync')
        self.logger.info('System installed')
        return True

    def destroy(self, *args, **kw):
        """
        Destroy the machine, and call super to release it in the inventory.
        """
        # TODO: Remove the machine from cobbler
        # TODO: Consider actually wrecking the install
        super(CobblerMachine, self).destroy(*args, **kw)
        self.__del__()
        del self

    def _cobble(self, cmd):
        """
        Pull out all cobbler commands so that we can later support changing
        users, remote login, not sudo, etc.
        """
        if self._runargs(['sudo', 'cobbler'] + cmd) != 0:
            error_msg = 'Cobbler command failed: ' + ' '.join(cmd)
            self.logger.error(error_msg)
            raise UTAHBMProvisioningException(error_msg)

    def _start(self):
        """
        Use cobbler to start the machine.
        """
        self.logger.debug('Starting system')
        self._cobble(['system', 'poweron', '--name=' + self.name])
        self.active = True

    def stop(self):
        """
        Use cobbler to stop the machine.
        """
        self.logger.debug('Stopping system')
        self._cobble(['system', 'poweroff', '--name=' + self.name])
        self.active = False

    def restart(self):
        """
        Stop the machine, wait powertimeout, then start it/
        """
        self.logger.info('Restarting system')
        self.stop()
        self.logger.debug('Waiting ' + str(config.powertimeout) + ' seconds')
        time.sleep(config.powertimeout)
        self._start()

    def __del__(self):
        """
        If the machine has an inventory, tell the inventory to release the
        machine when the object is deleted.
        If we exported a directory over NFS, remove it.
        """
        if self.installtype == 'desktop':
            self._removenfs()
        if self.inventory is not None:
            self.inventory.release(machine=self)

    def _removenfs(self):
        self.logger.info('Removing NFS share')
        self._runargs(['sudo', 'sed', '/' + self.tmpdir.replace('/', '\/') +
                       '/d', '-i', config.nfsconfigfile])
        self.logger.debug('Reloading NFS config')
        self._runargs(config.nfscommand)

    def activecheck(self):
        """
        Start the machine if needed, and check for SSH login.
        """
        # TODO: Consider moving this to SSHMixin
        self.logger.debug('Checking if machine is active')
        self.provisioncheck()
        if not self.active:
            self._start()
            self.sshcheck()
