#!/usr/bin/python

import os
from utah.provisioning.inventory.exceptions import \
        UTAHProvisioningInventoryException


class Inventory(object):
    """
    Provide a generic class for an arbitrary inventory of machines.

    Raise exceptions for most methods, since they should be defined by
    subclasses.
    Except for special read-only cases, subclasses should provide request().
    Subclasses will generally also need to provide delete() for cleanup
    purposes.
    All other methods (read, write, release, destroy) are optional.
    """
    def __init__(self, uniqueid, lockfile='~/.utah-inventory'):
        """
        To avoid provisioning machines to multiple requestors simultaneously,
        Inventory implements a lock file.
        Each inventory type should provide a unique id, which is combined with
        the class name to generate an inventory identifier.
        This identifier is stored in the lock file and checked on class
        instantiation to ensure the same inventory is running.
        Subclasses can circumvent this by not calling the superclass __init__
        function if necessary.
        """
        inventory = self.__class__.__name__ + str(uniqueid)
        self.lockfile = os.path.expanduser(lockfile)
        if os.path.isfile(self.lockfile):
            lockfile = open(self.lockfile, 'r')
            userinventory = lockfile.read()
            lockfile.close()
            if inventory != userinventory:
                raise UTAHProvisioningInventoryException(
                        'Different inventory in use: ' + userinventory
                        + "\nDelete " + self.lockfile
                        + ' if you are sure it is no longer valid')
        else:
            lockfile = open(self.lockfile, 'w')
            lockfile.write(inventory)
            lockfile.close()

    def delete(self):
        """
        Delete this inventory and remove the lockfile to facilitate use of
        another one.
        Subclasses can call this via super to delete the lockfile, or
        implement lockfile deletion in their delete() method.
        """
        os.unlink(self.lockfile)
        del self

    def read(self):
        """
        Read information about the inventory from any available sources.
        Primarily intended to force a synchronization of new external data if
        any exists.
        """
        return False

    def write(self):
        """
        Force a commit of all current inventory data in memory to internal or
        external storage.
        """
        return False

    def request(self, machinetype=None, *args, **kw):
        """
        Return a Machine object that can be provisioned and used to run tests.
        May interpret arguments and call the appropriate constructor with the
        appropriate arguments, or may take a Machine class as an argument and
        send all arguments to that constructor.
        """
        raise UTAHProvisioningInventoryException(
                'Method not defined for this inventory')

    def release(self):
        """
        Indicate that a machine is not now in use for tests, but may still have
        a working installed system, potentially suitable for further use.
        """
        return False

    def destroy(self):
        """
        Remove a machine from a inventory so that no further use of it will be
        requested or expected.
        Current implementations do not call machine.destroy(), requiring that
        do be done separately.
        """
        return False
