"""
Provide common functions for provisioning and interacting with machines.
Functions here should apply to multiple machine types (VM, bare metal, etc.)
"""

import logging
import logging.handlers
import os
import pipes
import shutil
import socket
import subprocess
import sys
import time
import urllib
import uuid
import paramiko
import re
import apt.cache

from utah.iso import ISO
from utah.preseed import Preseed
from utah.provisioning.exceptions import UTAHProvisioningException
from utah.retry import retry
import utah.timeout
from utah import config


class Machine(object):
    # TODO: make sure methods have consistent signatures
    # and return codes (or lack thereof)
    """
    Provide a generic class to provision an arbitrary machine.

    Raise exceptions for most methods, since subclasses should provide them.
    A fully implemented subclass will provide all public methods except read,
    I.E.:

    * provisioncheck, activecheck, getclientdeb, installclient
      (all these can be used from Machine or replaced.)
    * destroy, stop, uploadfiles, downloadfiles, run
      (these must be implemented separately.)
    """
    def __init__(self, arch=None, debug=False, directory=None, image=None,
                 dlpercentincrement=1, initrd=None, installtype=None,
                 kernel=None, machineid=None, machineuuid=None, name=None,
                 new=False, prefix='utah', preseed=None, series=None,
                 template=None, xml=None):
        """
        Initialize the object representing the machine.
        One of these groups of arguments should be included:
            series, installtype, arch: Download an ISO from the mirrors and use
                it to install the machine.
            image, kernel, initrd, preseed: Install the machine from a
                specified image/ISO file, with an optional kernel, initrd, and
                preseed.  libvirt xml file can be passed in xml as well.
            template: Clone the machine from a template or existing machine.
            name: Request a specific machine.  Combine with other groups to
                reinstall a specific machine.
        The subclass is responsible for provisioning with these arguments.
        Other arguments:
            debug: Enable debug logging.
            directory: Where the machine's files go.  Should be persistent for
                VMs, temporary is fine for installation-only files, like those
                used to install a physical machine using another method.
            dlpercentincrement: How often to log download updates to INFO
                (All updates logged to DEBUG)
            machineid: Should be passed by the request method of an Inventory
                class.  Often used to name virtual machines.
            machineuuid: Unique identifier.  Will be randomly generated if not
                supplied.
            new: Request a new machine (or a reinstall if a specific machine
                was requested.)
            prefix: prefix for automatically named machines.
        """
        # TODO: Make this work right with super at some point.
        self.arch = arch
        self.debug = debug
        self.dlpercentincrement = dlpercentincrement
        self.installtype = installtype
        self.machineid = machineid
        self.new = new
        self.prefix = prefix
        self.series = series
        self.template = template

        self._namesetup(name)
        self._dirsetup(directory)

        if machineuuid is None:
            self.uuid = str(uuid.uuid4())
        else:
            self.uuid = machineuuid

        self.provisioned = False
        self.active = False
        self._loggersetup()

        if image is None:
            self.image = None
        else:
            self.image = ISO(image,
                    dlpercentincrement=self.dlpercentincrement,
                    logger=self.logger)

        for item in ['preseed', 'xml', 'kernel', 'initrd']:
            # Ensure every file/url type argument is available locally
            arg = locals()[item]
            if arg is None:
                setattr(self, item, None)
            else:
                if arg.startswith('~'):
                    path = os.path.expanduser(arg)
                    self.logger.debug('Rewriting ~-based path: ' + arg
                                      + ' to: ' + path)
                else:
                    path = arg

                self.percent = 0
                self.logger.info('Preparing ' + item + ': ' + path)
                filename = urllib.urlretrieve(path,
                                              reporthook=self.dldisplay)[0]
                setattr(self, item, filename)
                self.logger.debug(path + ' is locally available as '
                                  + getattr(self, item))

        self.logger.debug('Machine init finished')

    def _namesetup(self, name=None):
        """
        Name the machine, automatically or using a specified name.
        """
        if name is None:
            self.name = self._makename()
        else:
            self.name = name

    def _dirsetup(self, directory=None):
        """
        Setup the machine's directory, automatically or using a specified one.
        """
        if directory is None:
            self.directory = os.path.join(config.vmpath, self.name)
        else:
            self.directory = directory

    def _makename(self, arch=None, installtype=None, machineid=None,
                  prefix=None, series=None):
        """
        Generate a name for the machine based on how it will be installed.
        Generally used for VMs.
        """
        if arch is None:
            arch = str(self.arch)
        if installtype is None:
            installtype = str(self.installtype)
        if machineid is None:
            machineid = str(self.machineid)
        if prefix is None:
            prefix = str(self.prefix)
        if series is None:
            series = str(self.series)
        if machineid is not None:
            prefix += ('-' + str(machineid))
        self.prefix = prefix
        if installtype == 'desktop':
            name = '-'.join([prefix, series, arch])
        else:
            name = '-'.join([prefix, series, installtype, arch])
        return name

    def _loggersetup(self):
        """
        Initialize the logging for the machine.
        Subclasses can override or supplement to customize logging.
        """
        self.logger = logging.getLogger('utah.provisioning.' + str(self.name))
        self.logger.propagate = False
        self.logger.setLevel(logging.INFO)
        for handler in list(self.logger.handlers):
            self.logger.removeHandler(handler)
            del handler
        self.consolehandler = logging.StreamHandler(stream=sys.stderr)
        console_formatter = logging.Formatter('%(levelname)s: %(message)s')
        self.consolehandler.setFormatter(console_formatter)
        self.consolehandler.setLevel(config.consoleloglevel)
        self.logger.addHandler(self.consolehandler)
        self.filehandler = logging.handlers.WatchedFileHandler(config.logfile)
        file_formatter = logging.Formatter('%(asctime)s '
                                           + self.name
                                           + ' %(levelname)s: %(message)s')
        self.filehandler.setFormatter(file_formatter)
        self.filehandler.setLevel(config.fileloglevel)
        self.logger.addHandler(self.filehandler)
        if config.debuglog is not None:
            self.logger.setLevel(logging.DEBUG)
            self.debughandler = (
                    logging.handlers.WatchedFileHandler(config.debuglog))
            self.debughandler.setFormatter(file_formatter)
            self.debughandler.setLevel(logging.DEBUG)
            self.logger.addHandler(self.debughandler)
        if self.debug:
            self.logger.setLevel(logging.DEBUG)

    def _loggerunsetup(self):
        """
        Remove logging.
        Principally used when a machine changes name after init.
        """
        del self.logger
        del self.consolehandler
        del self.filehandler

    def provisioncheck(self):
        """
        Check if the machine is provisioned, provision it if necessary, and
        raise an exception if it cannot be provisioned, or if it exceeds the
        timeout value set in config for installation.
        """
        self.logger.debug('Checking if machine is provisioned')
        if not self.provisioned:
            try:
                utah.timeout.timeout(config.installtimeout, self._provision)
            except utah.timeout.UTAHTimeout:
                raise UTAHProvisioningException('Machine install timed out.  '
                        + 'Consider adjusting timeout values in '
                        + '/etc/utah/conf, ~/.utah/conf, or '
                        + '$UTAH_CONFIG_FILE.')

    def activecheck(self):
        """
        Check if the machine is running and able to accept commands, start it
        if necessary, and raise an exception if it cannot be started.
        """
        self.logger.debug('Checking if machine is active')
        self.provisioncheck()
        if not self.active:
            self._start()

    def getclientdeb(self):
        """
        Return the path of the .deb file for the utah-client installed as part
        of the main package.
        """
        utah_version = apt.cache.Cache()['utah'].installedVersion
        if utah_version is None:
            raise UTAHProvisioningException("UTAH package isn't installed")

        basename = ('utah-client_' + utah_version + '_all.deb')
        debpath = os.path.join('/usr', 'share', 'utah', basename)
        debpath = os.path.normpath(debpath)
        if not os.path.isfile(debpath):
            raise UTAHProvisioningException('UTAH client binary package file '
                                            'not found in {path}\n'
                                            .format(path=debpath))
        self.logger.debug('Client deb path is ' + debpath)
        return debpath

    def installclient(self):
        """
        Install the utah-client package on the
        machine, and raise an exception if this fails.
        """
        self.logger.info('Installing client deb on machine')
        tmppath = os.path.normpath('/tmp')
        clientdeb = self.getclientdeb()
        try:
            self.uploadfiles([clientdeb], tmppath)
        except UTAHProvisioningException as err:
            try:
                for myfile in err.files:
                    if 'utah-client' in myfile:
                        raise UTAHProvisioningException('Client file not '
                                + 'found; UTAH was probably updated '
                                + 'during the run.  Please try again.',
                                retry=True)
                raise err
            except AttributeError:
                raise err
        if self.run('export DEBIAN_FRONTEND=noninteractive ; while '
                 + '(fuser /var/lib/dpkg/lock >/dev/null 2>&1) ; do echo '
                 + '"Waiting for dpkg lock to become available" ; sleep 10 ; '
                 + 'done ; dpkg -i ' + os.path.join(tmppath,
                 os.path.split(self.getclientdeb())[1])
                 + ' || apt-get -y -f install', root=True) != 0:
            raise UTAHProvisioningException('Failed to install client')

        remote_clientdeb = os.path.join(tmppath, os.path.basename(clientdeb))
        install_command = ('DEBIAN_FRONTEND=noninteractive '
                           'gdebi -n -q {}'
                           .format(remote_clientdeb))
        returncode, stdout, stderr = self.run(install_command, root=True)
        if (returncode != 0
            or re.search(r'script returned error exit status \d+', stderr)):
            raise UTAHProvisioningException('Failed to install client\n')

    def _provision(self):
        """
        Ready the machine for use, and set provisioned to True.
        If new=True or the machine is not installed, this should install the
        machine, possibly by calling _create().
        If an existing machine is requested, this should make the machine
        available, possibly by calling _load().
        Should generally not be called directly outside of the class;
        provisioncheck() or activecheck() should be used.
        """
        self._unimplemented(sys._getframe().f_code.co_name)

    def _create(self):
        """
        Install the system on the machine, and return True on success.
        Can take arguments if provided by provisioncheck(), or can read
        internal variables populated by __init__().
        Should generally not be called directly outside of the class;
        provisioncheck() or activecheck() should be used.
        """
        self._unimplemented(sys._getframe().f_code.co_name)

    def destroy(self):
        """
        Release resources consumed by the machine, set provisioned to False,
        and return True on success.
        For a VM, this should destroy the VM files on the disk, and remove it
        from libvirt if it is registered there.
        For a physical machine, this should free it up to be used by another
        process in the future, and destroy sensitive data if any exists.
        Destroying the install on a physical machine is optional, but it
        should be powered off if remote power management is available.
        Now returns False by default to work with multiple inheritance.
        """
        return False

    def _load(self):
        """
        Prepare an existing machine for use, and return True on success.
        Should generally not be called directly outside of the class;
        provisioncheck() or activecheck() should be used.
        """
        self._unimplemented(sys._getframe().f_code.co_name)

    def _start(self):
        """
        Start the machine, and set active to True.
        Should generally not be called directly outside of the class;
        activecheck() should be used.
        """
        self._unimplemented(sys._getframe().f_code.co_name)

    def stop(self, force=False):
        """
        Stop the machine, and set active to False.
        Graceful shutdown should be attempted if supported and force is False.
        A True value in force should ensure the machine is shutdown or raise
        an exception.
        """
        self._unimplemented(sys._getframe().f_code.co_name)

    def uploadfiles(self, files, target=os.path.normpath('/tmp/')):
        """
        Upload a list of local files to a target on the machine and return
        True if all uploads succeed.
        Support for files as a string specifying a single file is recommended
        but not required.
        target should be a valid target path for cp, i.e. a filename for a
        single file, a directory name for multiple files.
        """
        self._unimplemented(sys._getframe().f_code.co_name)

    def downloadfiles(self, files, target=os.path.normpath('/tmp/')):
        """
        Download a list of files from the machine to a local target and return
        True if all downloads succeed.
        Support for files as a string specifying a single file is recommended
        but not required.
        target should be a valid target path for cp, i.e. a filename for a
        single file, a directory name for multiple files.
        Recursive directory download is not currently supported by all
        implementations.
        """
        self._unimplemented(sys._getframe().f_code.co_name)

    def run(self, command, quiet=None, root=False, timeout=None):
        """
        Run a command on the machine and return a three element tuple:
        (returncode, stdout, stder)
        If quiet is True, all output other than the output of the command
        should be suppressed (i.e., no output from helper programs such as
        ssh.)
        If root is true, the command should be executed with elevated
        priveleges.
        If timeout is not None, it should be treated as a value in seconds
        after which run() will abort and return an error if no response was
        received.
        """
        self._unimplemented(sys._getframe().f_code.co_name)

    def read(self):
        """
        Load information about the machine into the machine object.
        This can be install information like series, arch, etc., or machine
        specific information like IP/MAC address, SSH key, or location of VM
        files or physical hardware.
        This method is optional.
        """
        return False

    def dircheck(self):
        self.logger.info('Checking if machine directory '
                         + self.directory + 'exists')
        if not os.path.isdir(self.directory):
            self.logger.debug
            os.makedirs(self.directory)
            self.logger.debug('Creating ' + self.directory)
        return True

    def dldisplay(self, blocks, size, total):
        read = blocks * size
        percent = 100 * read / total
        if percent >= self.percent:
            self.logger.info('File ' + str(percent) + '% downloaded')
            self.percent += self.dlpercentincrement
        self.logger.debug("%s read, %s%% of %s total" % (read, percent, total))

    def _runargs(self, arglist):
        """
        Pull command running out of individual classes to ease future
        modification.
        """
        self.logger.debug('Running command: ' + ' '.join(arglist))
        self.logger.debug('Output follows:')
        p = subprocess.Popen(arglist,
                             stdout=subprocess.PIPE, stderr=subprocess.STDOUT)
        while p.poll() is None:
            line = p.stdout.readline().strip()
            self.logger.debug(line)
        return p.returncode

    def _unimplemented(self, method):
        raise UTAHProvisioningException(self.__class__.__name__ +
              ' attempted to call the ' + method +
              ' method of the base Machine class, which is not implemented')


class SSHMixin(object):
    """
    Provide common commands for machines accessed via ssh.
    """
    def __init__(self, *args, **kwargs):
        # Note: Since this is a mixin it doesn't expect any argument
        # However, it calls super to initialize any other mixins in the mro
        super(SSHMixin, self).__init__(*args, **kwargs)
        ssh_client = paramiko.SSHClient()
        ssh_client.set_missing_host_key_policy(paramiko.AutoAddPolicy())
        self.ssh_client = ssh_client

    def run(self, command, quiet=None, root=False, timeout=None):
        """
        Run a command using ssh.
        """
        if isinstance(command, basestring):
            commandstring = command
        else:
            commandstring = ' '.join(command)
        if root:
            user = 'root'
        else:
            user = config.user

        self.activecheck()
        # Some commands expect run to return the output status of the command
        # We're going to try the method described here:
        # http://stackoverflow.com/questions/3562403/
        # With additions from here:
        # http://od-eon.com/blogs/
        # stefan/automating-remote-commands-over-ssh-paramiko/
        self.logger.debug('Connecting SSH')
        self.ssh_client.connect(self.name,
                                username=user,
                                key_filename=config.sshprivatekey)

        self.logger.debug('Opening SSH session')
        channel = self.ssh_client.get_transport().open_session()

        self.logger.info('Running command through SSH: ' + commandstring)
        stdout = channel.makefile('rb')
        stderr = channel.makefile_stderr('rb')
        if timeout is None:
            channel.exec_command(commandstring)
        else:
            utah.timeout.timeout(timeout, channel.exec_command, commandstring)
        retval = channel.recv_exit_status()

        self.logger.debug('Closing SSH connection')
        self.ssh_client.close()

        log_message = 'Return code: {}'.format(retval)
        if retval == 0:
            self.logger.debug(log_message)
        else:
            self.logger.warning(log_message)

        self.logger.debug('Standard output follows:')
        stdout_lines = stdout.readlines()
        for line in stdout_lines:
            self.logger.debug(line.strip())

        self.logger.debug('Standard error follows:')
        stderr_lines = stderr.readlines()
        for line in stderr_lines:
            self.logger.debug(line.strip())

        return retval, ''.join(stdout_lines), ''.join(stderr_lines)

    def uploadfiles(self, files, target=os.path.normpath('/tmp/')):
        """
        Copy a file or list of files to a target directory on the machine.
        """
        if isinstance(files, basestring):
            files = [files]

        self.activecheck()
        self.ssh_client.connect(self.name,
                                key_filename=config.sshprivatekey)
        sftp_client = self.ssh_client.open_sftp()
        failed = []
        try:
            for localpath in files:
                if os.path.isfile(localpath):
                    self.logger.info('Uploading ' + localpath
                                     + ' from the host to ' + target
                                     + ' on the machine')
                    remotepath = os.path.join(target,
                                              os.path.basename(localpath))
                    sftp_client.put(localpath, remotepath)
                else:
                    failed.append(localpath)
        finally:
            sftp_client.close()
            if len(failed) > 0:
                err = UTAHProvisioningException('Files do not exist: '
                                                + ' '.join(failed))
                err.files = failed
                raise err

    def downloadfiles(self, files, target=os.path.normpath('/tmp/')):
        """
        Copy a file or list of files from the machine to a target directory on
        the local system.
        """
        # TODO: check for directories and recurse into them
        if isinstance(files, basestring):
            files = [files]

        self.activecheck()
        self.ssh_client.connect(self.name,
                                key_filename=config.sshprivatekey)
        sftp_client = self.ssh_client.open_sftp()
        if os.path.isdir(target):
            get_localpath = lambda remotepath: \
                    os.path.join(target, os.path.basename(remotepath))
        else:
            get_localpath = lambda remotepath: target

        try:
            for remotepath in files:
                localpath = get_localpath(remotepath)
                self.logger.info('Downloading ' + remotepath
                                 + ' from the machine to ' + target
                                 + ' on the host')
                sftp_client.get(remotepath, localpath)
        finally:
            sftp_client.close()

    def downloadfilesindir(self, dirname, target=os.path.normpath('/tmp/')):
        """
        Copy all files in dirname on the machine to target on the host.
        Recursion is not yet implemented.
        """
        # TODO: Implement recursion
        self.activecheck()
        self.ssh_client.connect(self.name,
                                key_filename=config.sshprivatekey)
        sftp_client = self.ssh_client.open_sftp()
        files = [os.path.join(dirname, myfile)
                 for myfile in sftp_client.listdir(dirname)]
        self.downloadfiles(files, target)

    def destroy(self, *args, **kw):
        """
        Clean up known hosts for machine.
        """
        # TODO: evaluate value of known_hosts with paramiko
        self.logger.info('Removing machine addresses '
                         'from ssh known_hosts file')
        addresses = [self.name]
        try:
            addresses.append(socket.gethostbyname(self.name))
        except socket.gaierror as err:
            if err.errno in [-2, -5]:
                self.logger.debug(self.name
                                  + ' is not resolvable, '
                                  + 'so not removing from known_hosts')
            else:
                raise err

        old_host_keys = self.ssh_client.get_host_keys()
        new_host_keys = paramiko.HostKeys()
        addresses = set(addresses)
        for address, key in old_host_keys.iteritems():
            # Skip keys so that they don't get added
            # into the new keys (i.e. they're removed)
            if address in addresses:
                continue
            new_host_keys[address] = key
        new_host_keys.save(config.sshknownhosts)
        self.ssh_client.close()

        super(SSHMixin, self).destroy(*args, **kw)

    def sshcheck(self, timeout=config.checktimeout):
        """
        Sleep for a while and check if the machine is available via ssh.
        Return a retryable exception if it is not.
        Intended for use with retry.
        """
        self.logger.info('Checking for ssh availability')
        self.logger.info('Sleeping ' + str(timeout) + ' seconds')
        time.sleep(timeout)
        self.logger.info('Attempting to login')
        try:
            self.ssh_client.connect(self.name,
                                    key_filename=config.sshprivatekey)
        except socket.error as err:
            raise UTAHProvisioningException(str(err), retry=True)

    def sshpoll(self, timeout=config.boottimeout,
                checktimeout=config.checktimeout, logmethod=None):
        """
        Run sshcheck over and over until timeout expires.
        """
        if logmethod is None:
            logmethod = self.logger.debug
        utah.timeout.timeout(timeout, retry, self.sshcheck, checktimeout,
                             logmethod=logmethod)


class CustomInstallMixin(object):
    """
    Provide routines for unpacking necessary boot files from images,
    rewriting them to our liking, and repacking them.
    """
    # TODO: setup variable(s) to suppress adding and rewriting
    def _preparekernel(self, kernel=None, tmpdir=None):
        """
        If we haven't been given a kernel file, unpack one from the image.
        """
        # TODO: document this better
        # TODO: consider moving this into the ISO class when it exists
        self.logger.info('Preparing kernel')
        if kernel is None:
            kernel = self.kernel
        if tmpdir is None:
            tmpdir = self.tmpdir
        if kernel is None:
            self.logger.info('Unpacking kernel from image')
            kernelpath = './install/vmlinuz'
            if self.installtype == 'mini':
                self.logger.debug('Mini image detected')
                kernelpath = 'linux'
            elif self.installtype == 'desktop':
                self.logger.debug('Desktop image detected')
                kernelpath = './casper/vmlinuz'
            kernel = os.path.join(tmpdir, 'kernel')
            self.image.extract(kernelpath, outfile=kernel)
        else:
            self.logger.info('Using local kernel: ' + kernel)
        return kernel

    def _prepareinitrd(self, initrd=None, tmpdir=None):
        """
        If we haven't been given an initrd file, unpack one from the image.
        """
        # TODO: document this better
        # TODO: consider moving this into the ISO class when it exists
        self.logger.info('Preparing initrd')
        if initrd is None:
            initrd = self.initrd
        if tmpdir is None:
            tmpdir = self.tmpdir
        if initrd is None:
            self.logger.info('Unpacking initrd from image')
            initrdpath = './install/initrd.gz'
            if self.installtype == 'mini':
                self.logger.debug('Mini image detected')
                initrdpath = 'initrd.gz'
            elif self.installtype == 'desktop':
                self.logger.debug('Desktop image detected')
                initrdpath = './casper/initrd.lz'
            initrd = os.path.join(tmpdir, os.path.basename(initrdpath))
            self.image.extract(initrdpath, outfile=initrd)
        else:
            self.logger.info('Using local initrd: ' + initrd)
        return initrd

    def _unpackinitrd(self, initrd=None, tmpdir=None):
        """
        Unpack the initrd file into a directory so we can modify it.
        """
        self.logger.info('Unpacking initrd')
        if initrd is None:
            initrd = self.initrd
        if tmpdir is None:
            tmpdir = self.tmpdir
        if not os.path.isdir(os.path.join(tmpdir, 'initrd.d')):
            os.makedirs(os.path.join(tmpdir, 'initrd.d'))
        os.chdir(os.path.join(tmpdir, 'initrd.d'))
        pipe = pipes.Template()
        if os.path.splitext(initrd)[1] == '.gz':
            self.logger.debug('Using gzip based on file extension')
            pipe.prepend('zcat $IN', 'f-')
        elif os.path.splitext(initrd)[1] == '.lz':
            self.logger.debug('Using lzma based on file extension')
            pipe.prepend('lzcat -S .lz $IN', 'f-')
        pipe.append('cpio -ivd 2>/dev/null', '-.')
        pipe.copy(initrd, '/dev/null')

    def _setuplatecommand(self, tmpdir=None):
        """
        Setup the latecommand script we run during a preseeded install.
        """
        # TODO: document this better
        self.logger.info('Copying ssh public key')
        if tmpdir is None:
            tmpdir = self.tmpdir
        shutil.copyfile(config.sshpublickey,
                        os.path.join(tmpdir, 'initrd.d', 'utah-ssh-key'))

        self.logger.info('Creating latecommand script')
        filename = os.path.join(tmpdir, 'initrd.d', 'utah-latecommand')
        latecommand = open(filename, 'w')
        latecommand.write(
"""#!/bin/sh

ROOTDIR=/target/root/.ssh
USERDIR=/target/home/{user}/.ssh

for DIR in $ROOTDIR $USERDIR
do
  mkdir -p $DIR
  cp utah-ssh-key $DIR/authorized_keys
  chmod 700 $DIR
done

chown -R root:root $ROOTDIR
chown -R 1000:1000 $USERDIR

mkdir -p /target/etc/utah
echo "{uuid}" > /target/etc/utah/uuid
echo "{user}" >> /target/etc/utah/users

# Add stable PPA to get latest python-jsonschema version
chroot /target sh -c 'add-apt-repository -y ppa:utah/stable; apt-get update'
""".format(user=config.user, uuid=self.uuid))
        latecommand.close()

    def _setuppreseed(self, tmpdir=None):
        """
        Rewrite the preseed to include what we need for an automated install
        and subsequent access.
        """
        # TODO: document this better
        # TODO: if the lines we need aren't in the preseed, add them
        self.logger.info('Setting up preseed')
        if tmpdir is None:
            tmpdir = self.tmpdir
        preseed = Preseed(self.preseed)
        if 'preseed/late_command' in preseed:
            question = preseed['preseed/late_command']
            if self.installtype == 'desktop':
                self.logger.info('Changing d-i latecommand '
                        + 'to ubiquity success_command '
                        + 'and prepending ubiquity lines')
                question.owner = 'ubiquity'
                question.qname = 'ubiquity/success_command'
                question.value.prepend(
                        'chroot /target sh -c \''        # chroot start
                        'export LOG_FILE=/var/log/utah-install; '
                        'apt-get install -y --force-yes openssh-server '
                        '>>$LOG_FILE 2>&1 ; '
                        'apt-get install -y --force-yes gdebi-core '
                        '>>$LOG_FILE 2>&1 ; '
                        '\'; '                           # chroot end
                        )
                question.prepend("ubiquity ubiquity/summary note")
                question.prepend("ubiquity ubiquity/reboot boolean true")
            else:
                self.logger.info('Prepending latecommand')
                question.value.prepend('sh utah-latecommand ; ')
        if 'ubiquity/success_command' in preseed:
            self.logger.info('Prepending success_command')
            question = preseed['ubiquity/success_command']
            question.value.prepend('sh utah-latecommand ; ')
        if 'pkgsel/include' in preseed:
            question = preseed['pkgsel/include']
            packages = question.value.text.split()
            for pkgname in ('openssh-server', 'gdebi-core'):
                if pkgname not in packages:
                    self.logger.info('Adding {} to preseeded packages'
                                     .format(pkgname))
                    packages.append(pkgname)
            question.value = ' '.join(packages)

        if 'netcfg/get_hostname' in preseed:
            self.logger.info('Rewriting hostname to ' + self.name)
            question = preseed['netcfg/get_hostname']
            question.value = self.name
        if 'passwd/username' in preseed:
            self.logger.info('Rewriting username to ' + config.user)
            question = preseed['passwd/username']
            question.value = config.user

        preseed.write(os.path.join(tmpdir, 'initrd.d', 'preseed.cfg'))

        if self.installtype == 'desktop':
            self.logger.info('Inserting preseed into casper')
            preseedscript = os.path.join(tmpdir, 'initrd.d', 'scripts',
                                         'casper-bottom', 'utah')
            open(preseedscript, 'w').write(
"""#!/bin/sh

cp preseed.cfg utah-ssh-key utah-latecommand /root
cp 50-utahdefault.conf /root/etc/rsyslog.d
""")
            os.chmod(preseedscript, 0755)
            casper_dir = os.path.join(tmpdir, 'initrd.d', 'scripts',
                                       'casper-bottom')
            tmpfilename = os.path.join(casper_dir, 'ORDER.tmp')
            tmpfile = open(tmpfilename, 'w')
            orderfilename = os.path.join(casper_dir, 'ORDER')
            orderfile = open(orderfilename, 'r')
            tmpfile.write(
"""/scripts/casper-bottom/utah
[ -e /conf/param.conf ] && . /conf/param.conf
""")
            tmpfile.flush()
            for line in orderfile:
                tmpfile.write(line)
                tmpfile.flush()
            tmpfile.close()
            orderfile.close()
            os.rename(tmpfilename, orderfilename)

    def _setuplogging(self, tmpdir=None):
        """
        Send the installer syslog to a console.
        """
        # TODO: move this into CustomVM
        if tmpdir is None:
            tmpdir = self.tmpdir
        inittab = os.path.join(tmpdir, 'initrd.d', 'etc', 'inittab')
        if os.path.isfile(inittab):
            self.logger.info('Updating inittab')
            with open(inittab, 'a') as myfile:
                myfile.write("\n"
                             "# logging to serial\n"
                             "ttyS0::respawn:/usr/bin/tail -n 1200 "
                             "-f /var/log/syslog \n")
        else:
            self.logger.info('No inittab found, creating rsyslog config file')
            conffilename = os.path.join(tmpdir, 'initrd.d',
                                        '50-utahdefault.conf')
            open(conffilename, 'w').write(
"""
#  Default rules for rsyslog.
#  Minimal file provided for utah installation logging
#  For more information see rsyslog.conf(5) and /etc/rsyslog.conf
auth,authpriv.*			/var/log/auth.log
*.*;auth,authpriv.none		-/dev/ttyS0
*.*;auth,authpriv.none		-/var/log/syslog
kern.*				-/var/log/kern.log
mail.*				-/var/log/mail.log
mail.err			/var/log/mail.err
news.crit			/var/log/news/news.crit
news.err			/var/log/news/news.err
news.notice			-/var/log/news/news.notice
*.emerg				*
daemon.*;mail.*;\
	news.err;\
	*.=debug;*.=info;\
	*.=notice;*.=warn	|/dev/xconsole
""")

    def _repackinitrd(self, tmpdir=None):
        """
        Pack an initrd from our directory.
        """
        self.logger.info('Repacking initrd')
        if tmpdir is None:
            tmpdir = self.tmpdir
        pipe = pipes.Template()
        pipe.prepend('find .', '.-')
        pipe.append('cpio --quiet -o -H newc', '--')
        # Desktop image loads initrd.gz,
        # but for physical machines we should stick with lz
        if self.installtype == 'desktop':
            self.logger.debug('Using lzma because installtype is desktop')
            pipe.append('lzma -9fc ', '--')
            initrd = os.path.join(tmpdir, 'initrd.lz')
        else:
            self.logger.debug('Using gzip because installtype is not desktop')
            pipe.append('gzip -9fc ', '--')
            initrd = os.path.join(tmpdir, 'initrd.gz')
        pipe.copy('/dev/null', initrd)
        return initrd

    def _cmdlinesetup(self, boot=None):
        """
        Add the needed options to the command line for an automatic install.
        """
        if boot is None:
            self.cmdline = ''
        else:
            self.cmdline = boot
        self.boottimeout = config.boottimeout
        if self.installtype == 'desktop':
            if 'automatic-ubiquity' not in self.cmdline:
                self.logger.info('Adding automatic-ubiquity option '
                                 'since installtype is desktop')
                self.cmdline += ' automatic-ubiquity'
            if 'boot=' not in self.cmdline:
                self.logger.info('Adding boot=casper option '
                                 'since installtype is desktop')
                self.cmdline += ' boot=casper'
            if 'keyboard-configuration/layoutcode=' not in self.cmdline:
                self.logger.info('Adding keyboard-configuration/layoutcode=us '
                                 'option since installtype is desktop')
                self.cmdline += ' keyboard-configuration/layoutcode=us'
            if 'noprompt' not in self.cmdline:
                self.logger.info('Adding noprompt option '
                                 'since installtype is desktop')
                self.cmdline += ' noprompt'
        else:
            if (self.installtype == 'server'
                and 'DEBCONF_DEBUG' not in self.cmdline):
                self.logger.info('Adding DEBCONF_DEBUG=developer option '
                                 'since installtype is not server')
                self.cmdline += ' DEBCONF_DEBUG=developer'
            if 'debconf/priority=' not in self.cmdline:
                self.logger.info('Adding debconf/priority=critical option '
                                 'since installtype is not desktop')
                self.cmdline += ' debconf/priority=critical'
        self.cmdline.strip()

    def _custominit(self, arch=None, boot=None, installtype=None, series=None):
        """
        Setup installtype, arch, series and commandline.
        Everything but commandline setup may be replaced by ISO class later.
        """
        #TODO: Make this a proper __init__ method
        #by adding super call to Machine.
        if installtype is None:
            self.installtype = self.image.installtype
        if arch is None:
            self.arch = self.image.arch
        if series is None:
            self.series = self.image.series
        self._cmdlinesetup(boot=boot)
