"""
Provide classes to provision libvirt-based virtual machines.
"""

import os
import random
import shutil
import string
import subprocess
import tempfile

from xml.etree import ElementTree

import apt.cache
import libvirt

from utah.provisioning.provisioning import (
        SSHMixin,
        CustomInstallMixin,
        )
from utah.provisioning.vm.vm import VM
from utah.provisioning.vm.exceptions import UTAHVMProvisioningException
from utah import config
from utah.process import ProcessChecker


class LibvirtVM(VM):
    """
    Provide a class to utilize VMs using libvirt.

    Capable of utilizing existing VMs.
    Creation currently handled by sublcasses.
    """
    def __init__(self, *args, **kw):
        super(LibvirtVM, self).__init__(*args, **kw)
        libvirt.registerErrorHandler(self.libvirterrorhandler, None)
        self.lv = libvirt.open(config.qemupath)
        if self.lv is None:
            raise UTAHVMProvisioningException('Cannot connect to libvirt')
        self.logger.debug('LibvirtVM init finished')

    def _load(self):
        """
        Load an existing VM.
        """
        self.logger.info('Loading VM')
        self.vm = self.lv.lookupByName(self.name)
        self.logger.info('VM loaded')
        return True

    def _provision(self):
        """
        Make an existing VM available using libvirt to look up the VM by name.
        """
        self.logger.info('Provisioning VM')
        if self.new:
            self.logger.debug('New VM requested')
            try:
                self._load()
                self.logger.error('VM already exists')
                raise UTAHVMProvisioningException('Request new VM, but '
                                                  + self.name
                                                  + ' already exists')
            except libvirt.libvirtError as err:
                if err.get_error_code() == 42:
                    self._create()
                else:
                    raise err

        try:
            self._load()
        except libvirt.libvirtError as err:
            if err.get_error_code() == 42:
                self.logger.debug('Lookup failed')
                try:
                    self._create()
                    self._load()
                except UTAHVMProvisioningException as error:
                    self.logger.error('VM lookup failed')
                    raise UTAHVMProvisioningException('Cannot find VM named '
                                                      + self.name +
                                                      ' and ' + str(error))
            else:
                raise err
        self.provisioned = True
        self.logger.info('VM provisioned')

    def activecheck(self):
        """
        Verify the machine is provisioned, then start it if it is not started.
        """
        self.logger.debug('Checking if VM is active')
        self.provisioncheck()
        if self.vm.isActive() == 0:
            self._start()
        else:
            self.active = True

    def _start(self):
        """
        Verify the machine is provisioned, then start it.
        """
        self.logger.info('Starting VM')
        self.vm.create()
        self.active = True

    def stop(self, force=False):
        """
        Stop the machine.
        Setting force to true will do a hard shutdown instead of a graceful
        one.
        """
        self.logger.info('Stopping VM')
        if self.vm.isActive() == 0:
            self.logger.info('VM is already stopped')
        else:
            if force:
                self.logger.info('Forced shutdown requested')
                self.vm.destroy()
            else:
                self.vm.shutdown()
        self.active = False

    def libvirterrorhandler(self, context, err):
        """
        Log libvirt errors instead of sending them directly to the console.
        """
        errorcode = err.get_error_code()
        if errorcode in [9, 42]:
            # We see these as part of normal operations,
            # so we send them to debug
            # 9 is trying to create a VM that already exists
            # 42 is trying to load a VM that doesn't exist
            logmethod = self.logger.debug
        else:
            logmethod = self.logger.error
        logmethod('libvirt error: ' + err['message'])
        logmethod('libvirt error number is: ' + str(errorcode))


class VMToolsVM(SSHMixin, LibvirtVM):
    """
    Provide a class to provision a VM using the ubuntu-qa-tools vm-tools.
    """
    def __init__(self, arch='i386', installtype='desktop', machineid=None,
                 prefix='utah', series='precise', *args, **kw):
        if not apt.cache.Cache()['vm-tools'].is_installed:
            raise UTAHVMProvisioningException('vm-tools is not installed.  '
                    'Try: sudo apt-get install vm-tools')
        super(VMToolsVM, self).__init__(*args, arch=arch,
                                        installtype=installtype,
                                        machineid=machineid,
                                        name=None,
                                        series=series,
                                        **kw)
        self.logger.debug('VMToolsVM init finished')

    def _start(self):
        """
        Start the VM using vm-start, which will wait until SSH is up.
        """
        self.logger.info('Starting the vm using vm-start')
        args = ['vm-start', '-v', '-w', self.name]
        self.active = (self._runargs(args) == 0)
        return self.active

    def activecheck(self):
        """
        Verify the machine is provisioned, then start it if it is not started.
        Use vm-wait to make sure it's up.
        """
        self.provisioncheck()
        self.logger.debug('Checking if VMToolsVM is active')
        self._start()
        self.logger.info('Using vm-wait to ensure VM is active')
        if (self._runargs(['vm-wait', self.name, '300']) != 0):
            self.active = False
            raise UTAHVMProvisioningException('Timed out waiting for VM '
                                              'to be reachable')
        else:
            self.active = True

    def _getcreationargs(self):
        """
        Return the vm-new syntax used to create the VM.
        Can be used for debugging or instructional purposes.
        """
        args = ['vm-new', '-f', '-r', '-v', '-b', config.bridge,
                '-t', self.installtype, self.series, self.arch, self.prefix]
        self.logger.debug('VM Creation args: ' + str(args))
        return args

    def _create(self):
        """
        Run the command generated by getcreationargs
        and either return a good status or raise an exception.
        """
        self.logger.info('Creating vm using vm-new')
        self.logger.info('This may take up to two hours, '
                         'excluding download times, and may go over an hour '
                         'without output during virt-install')
        args = self._getcreationargs()
        percent = 0
        p = subprocess.Popen(args,
                             stdout=subprocess.PIPE, stderr=subprocess.STDOUT)
        while p.poll() is None:
            line = p.stdout.readline().strip()
            self.logger.debug(line)
            if 'Fetching release iso' in line:
                self.logger.info('Downloading ISO')
            if len(line.split()) >= 3 and '%' in line.split()[-3]:
                dlpercent = int(line.split()[-3].strip('%'))
                if dlpercent >= percent:
                    self.logger.info('ISO ' + str(dlpercent) + '% downloaded, '
                                     + line.split(' ')[-1] + ' remaining')
                    percent += self.dlpercentincrement
            if ' saved ' in line:
                self.logger.info('ISO download complete')
            if 'Creating preseeded iso' in line:
                self.logger.info('Preseeding ISO')
            if 'virt-install' in line:
                self.logger.info('Installing system on VM '
                                 '(may take over an hour)')
                self.logger.info('You can watch the progress with:')
                self.logger.info("\tvm-view " + self.name)
                self.logger.info('Take care not to interrupt the install')
            if 'Verifying lsb_release' in line:
                self.logger.info('Waiting for post-install '
                                 'and verifying system '
                                 '(may take over a half hour)')
            if "No domains available for virt type 'hvm'" in line:
                self.logger.error('No hardware virtual machine '
                                  'support available')
                self.logger.info('Please ensure the following:')
                self.logger.info("\tYour processor supports hardware "
                                 "virtualization extensions")
                self.logger.info("\tHardware virtualization "
                                 "is not disabled in the BIOS")
                self.logger.info("\tkvm is installed")
                self.logger.info("\tThe kvm and processor-specific "
                                 "kvm kernel modules are installed and loaded")
                self.logger.error('Software virtual machine support '
                                  'is implemented in the CustomVM class')
                raise UTAHVMProvisioningException(
                        'No hardware virtual machine support available')
            if ('Could not find' in line
                and ".uqt-vm-tools.conf' configuration file!" in line):
                self.logger.error('No .uqt-vm-tools.conf configuration file '
                                  'found in home directory')
                self.logger.info('If you are running as the utah user, '
                                 'you can use the packaged config file:')
                self.logger.info("\tln -s /etc/utah/uqt-vm-tools.conf "
                                 "~utah/.uqt-vm-tools.conf")
                self.logger.info('As a different user, '
                                 'run vm-new with no arguments, '
                                 'and answer y when prompted '
                                 'to create an initial config file')
                self.logger.info('We recommend adding python-yaml '
                                 'to the vm_extra_packages line '
                                 'of this file, i.e.:')
                self.logger.info("\tsed "
                                 "'s/^vm_extra_packages=\""
                                 "/vm_extra_packages=\"python-yaml /' "
                                 "-i ~/.uqt_vm_tools.conf")
                raise UTAHVMProvisioningException(
                        'No vm-tools config file available; '
                        'more info available in '
                        + self.filehandler.baseFilename)
        if p.returncode == 0:
            return True
        else:
            raise UTAHVMProvisioningException(
                    'Failed to create VM: vm-new exit status: '
                    + str(p.returncode))

    def destroy(self):
        """
        Use vm-remove to destroy the vm.
        """
        self.logger.info('Destroying vm using vm-remove')
        self.provisioned = False
        args = ['vm-remove', '-f', self.name]
        return (self._runargs(args) == 0)

    def _getcommandargs(self, command, quiet=None, root=False, timeout=300):
        """
        Get the arguments to send to vm-cmd.
        Used by multiple other functions.
        """
        if quiet is None:
            quiet = not self.debug
        args = ['vm-cmd', '-f']
        if timeout is not None:
            args.extend(['-s', '-t', str(timeout)])
        if quiet:
            args.append('-q')
        if root:
            args.append('-r')
        args.append(self.name)
        if isinstance(command, basestring):
            args.append(command)
        else:
            args.extend(command)
        self.logger.debug('vm-cmd command args: ' + str(args))
        return args

    def run(self, command, quiet=None, root=False, timeout=300):
        """
        Run a command using vm-cmd.
        Timeout defaults to 300 seconds, but can be set to None to not send a
        timeout argument.
        """
        if quiet is None:
            quiet = not self.debug
        self.activecheck()
        args = self._getcommandargs(command=command, quiet=quiet,
                             root=root, timeout=timeout)
        self.logger.info('Running command on VM: ' + ' '.join(args))
        p = subprocess.Popen(args,
                             stdout=subprocess.PIPE, stderr=subprocess.PIPE)
        while p.poll() is None:
            pass
        return p.returncode, p.communicate()[0], p.communicate()[1]

    def uploadfiles(self, files, target=os.path.normpath('/tmp/')):
        """
        Copy a file or list of files to a target directory on the machine.
        """
        if isinstance(files, basestring):
            files = [files]
        if not target.endswith('/'):
            target += '/'
        self.activecheck()
        success = True
        for myfile in files:
            self.logger.info('Uploading ' + myfile
                             + ' from the host to ' + target + ' on the VM')
            returncode = self._runargs(['vm-scp', '-f', '-p',
                                        self.name, myfile, target])
            if returncode != 0:
                success = False
        return success


class CustomVM(CustomInstallMixin, SSHMixin, LibvirtVM):
    """
    Install a VM from an image using libvirt direct kernel booting.
    """
    def __init__(self, arch=None, boot=None, diskbus=None, disksizes=None,
                 emulator=None, image=None, initrd=None, installtype=None,
                 kernel=None, machineid=None, macs=[], name=None,
                 prefix='utah', preseed=None, series=None, xml=None,
                 *args, **kw):
        # Make sure that no other virtualization solutions are running
        # TODO: see if this is needed for qemu or just kvm
        process_checker = ProcessChecker()
        for cmdline, app in [('/usr/lib/virtualbox/VirtualBox', 'VirtualBox'),
                             ('/usr/lib/vmware/bin', 'VMware')]:
            if process_checker.check_cmdline(cmdline):
                message = process_checker.get_error_message(app)
                raise UTAHVMProvisioningException(message)

        if diskbus is None:
            self.diskbus = 'virtio'
        else:
            self.diskbus = diskbus
        if disksizes is None:
            self.disksizes = [8]
        else:
            self.disksizes = disksizes
        if name is None:
            autoname = True
            name = '-'.join([str(prefix), str(machineid)])
        else:
            autoname = False
        if preseed is None:
            preseed = '/etc/utah/default-preseed.cfg'
        if xml is None:
            xml = '/etc/utah/default-vm.xml'
        super(CustomVM, self).__init__(arch=arch, image=image, initrd=initrd,
                                       installtype=installtype, kernel=kernel,
                                       machineid=machineid, name=name,
                                       preseed=preseed, series=series, xml=xml,
                                       *args, **kw)
        # TODO: do a better job of separating installation
        # into _create rather than __init__
        if self.image is None:
            raise UTAHVMProvisioningException('Image file required '
                                              'for custom VM installation')
        self._custominit(arch=arch, boot=boot,
                         installtype=installtype, series=series)
        if autoname:
            self._namesetup()
            self._loggerunsetup()
            self._loggersetup()
            self._dirsetup()
        if emulator is None:
            if self._supportsdomaintype('kvm'):
                self.logger.info('Setting type to kvm '
                                 'since it is present in libvirt capabilities')
                self.domaintype = 'kvm'
            elif self._supportsdomaintype('qemu'):
                self.logger.info('Setting type to qemu '
                                 'since it is present in libvirt capabilities')
                self.domaintype = 'qemu'
            else:
                raise UTAHVMProvisioningException(
                        'kvm and qemu not supported in libvirt capabilities; '
                        'please make sure qemu and/or kvm are installed '
                        'and libvirt is configured correctly')
        else:
            self.domaintype = emulator
        if self.domaintype == 'qemu':
            self.logger.debug('Raising boot timeout for qemu domain')
            self.boottimeout *= 4
        self.macs = macs
        self.xml = xml
        self.dircheck()
        self.logger.debug('CustomVM init finished')

    def _createdisks(self, disksizes=None):
        """
        Create disk files if needed and build a list of them.
        """
        self.logger.info('Creating disks')
        if disksizes is None:
            disksizes = self.disksizes
        self.disks = []
        for index, size in enumerate(disksizes):
            disksize = '{}G'.format(size)
            basename = 'disk{}.qcow2'.format(index)
            diskfile = os.path.join(self.directory, basename)
            if not os.path.isfile(diskfile):
                cmd = ['qemu-img', 'create', '-f', 'qcow2', diskfile, disksize]
                self.logger.debug('Creating ' + disksize + ' disk using:')
                self.logger.debug(' '.join(cmd))
                if self._runargs(cmd) != 0:
                    raise UTAHVMProvisioningException(
                            'Could not create disk image at ' + diskfile)
            disk = {'bus': self.diskbus,
                    'file': diskfile,
                    'size': disksize,
                    'type': 'qcow2'}
            self.disks.append(disk)
            self.logger.debug('Adding disk to list')

    def _supportsdomaintype(self, domaintype):
        """
        Check emulator support in libvirt capabilities.
        """
        capabilities = ElementTree.fromstring(self.lv.getCapabilities())
        for guest in capabilities.iterfind('guest'):
            for arch in guest.iterfind('arch'):
                for domain in arch.iterfind('domain'):
                    if domaintype in domain.get('type'):
                        return True
        return False

    def _installxml(self, cmdline=None, initrd=None,
                    kernel=None, tmpdir=None, xml=None):
        """
        Return the XML tree to be passed to libvirt for VM installation.
        """
        self.logger.info('Creating installation XML')
        if cmdline is None:
            cmdline = self.cmdline
        if initrd is None:
            initrd = self.initrd
        if kernel is None:
            kernel = self.kernel
        if xml is None:
            xml = self.xml
        if tmpdir is None:
            tmpdir = self.tmpdir
        xmlt = ElementTree.ElementTree(file=xml)
        self.logger.debug('Rewriting basic info')
        xmlt.find('name').text = self.name
        xmlt.find('uuid').text = self.uuid
        self.logger.debug('Setting type to qemu '
                          'in case no hardware virtualization present')
        xmlt.getroot().set('type', self.domaintype)
        ose = xmlt.find('os')
        if self.arch == ('i386'):
            ose.find('type').set('arch', 'i686')
        elif self.arch == ('amd64'):
            ose.find('type').set('arch', 'x86_64')
        else:
            ose.find('type').set('arch', self.arch)
        self.logger.debug('Setting up boot info')
        for kernele in list(ose.iterfind('kernel')):
            ose.remove(kernel)
        kernele = ElementTree.Element('kernel')
        kernele.text = kernel
        ose.append(kernele)
        for initrde in list(ose.iterfind('initrd')):
            ose.remove(initrd)
        initrde = ElementTree.Element('initrd')
        initrde.text = initrd
        ose.append(initrde)
        for cmdline in list(ose.iterfind('cmdline')):
            ose.remove(cmdline)
        cmdlinee = ElementTree.Element('cmdline')
        cmdlinee.text = cmdline
        ose.append(cmdlinee)
        xmlt.find('on_reboot').text = 'destroy'
        self.logger.debug('Setting up devices')
        devices = xmlt.find('devices')
        self.logger.debug('Setting up disks')
        for disk in list(devices.iterfind('disk')):
            if disk.get('device') == 'disk':
                devices.remove(disk)
                self.logger.debug('Removed existing disk')
            #TODO: Add a cdrom if none exists
            if disk.get('device') == 'cdrom':
                if disk.find('source') is not None:
                    disk.find('source').set('file', self.image.image)
                    self.logger.debug('Rewrote existing CD-ROM')
                else:
                    source = ElementTree.Element('source')
                    source.set('file', self.image.image)
                    disk.append(source)
                    self.logger.debug('Added source to existing CD-ROM')
        for disk in self.disks:
            diske = ElementTree.Element('disk')
            diske.set('type', 'file')
            diske.set('device', 'disk')
            driver = ElementTree.Element('driver')
            driver.set('name', 'qemu')
            driver.set('type', disk['type'])
            diske.append(driver)
            source = ElementTree.Element('source')
            source.set('file', disk['file'])
            diske.append(source)
            target = ElementTree.Element('target')
            dev = "vd%s" % (string.ascii_lowercase[self.disks.index(disk)])
            target.set('dev', dev)
            target.set('bus', disk['bus'])
            diske.append(target)
            devices.append(diske)
            self.logger.debug('Added ' + str(disk['size']) + ' disk')
        macs = list(self.macs)
        for interface in devices.iterfind('interface'):
            if interface.get('type') in ['network', 'bridge']:
                if len(macs) > 0:
                    mac = macs.pop(0)
                    interface.find('mac').set('address', mac)
                    self.logger.debug('Rewrote interface '
                                      'to use specified mac address ' + mac)
                else:
                    mac = random_mac_address()
                    interface.find('mac').set('address', mac)
                    self.macs.append(mac)
                    self.logger.debug('Rewrote interface '
                                      'to use random mac address ' + mac)
            if interface.get('type') == 'bridge':
                interface.find('source').set('bridge', config.bridge)
        serial = ElementTree.Element('serial')
        serial.set('type', 'file')
        source = ElementTree.Element('source')
        log_filename = os.path.join(config.logpath, self.name + '.syslog.log')
        source.set('path', log_filename)
        serial.append(source)
        target = ElementTree.Element('target')
        target.set('port', '0')
        serial.append(target)
        devices.append(serial)
        if self.debug:
            xmlt.write(os.path.join(tmpdir, 'install.xml'))
        self.logger.info('Installation XML ready')
        return xmlt

    def _installvm(self, lv=None, tmpdir=None, xml=None):
        """
        Install a VM, then undefine it in libvirt.
        The final installation will recreate the VM using the existing disks.
        """
        self.logger.info('Creating VM')
        if lv is None:
            lv = self.lv
        if xml is None:
            xml = self.xml
        if tmpdir is None:
            tmpdir = self.tmpdir
        vm = lv.defineXML(ElementTree.tostring(xml.getroot()))
        os.chmod(tmpdir, 0755)
        vm.create()
        self.logger.info('Installing system on VM (may take over an hour)')
        self.logger.info('You can watch the progress with virt-viewer')
        log_filename = os.path.join(config.logpath, self.name + '.syslog.log')
        self.logger.info('Logs will be written to ' + log_filename)

        while vm.isActive() is not 0:
            pass

        vm.undefine()
        self.logger.info('Installation complete')

    def _finalxml(self, tmpdir=None, xml=None):
        """
        Create the XML to be used for the post-installation VM.
        This may be a transformation of the installation XML.
        """
        self.logger.info('Creating final VM XML')
        if xml is None:
            xml = ElementTree.ElementTree(file=self.xml)
        if tmpdir is None:
            tmpdir = self.tmpdir
        ose = xml.find('os')
        for kernel in ose.iterfind('kernel'):
            ose.remove(kernel)
        for initrd in ose.iterfind('initrd'):
            ose.remove(initrd)
        for cmdline in ose.iterfind('cmdline'):
            ose.remove(cmdline)
        xml.find('on_reboot').text = 'restart'
        devices = xml.find('devices')
        devices.remove(devices.find('serial'))
        if self.debug:
            xml.write(os.path.join(tmpdir, 'final.xml'))
        return xml

    def _create(self):
        """
        Create the VM, install the system, and prepare it to boot.
        This primarily calls functions from CustomInstallMixin and CustomVM.
        """
        self.logger.info('Creating custom virtual machine')

        tmpdir = tempfile.mkdtemp(prefix='/tmp/' + self.name + '_')
        self.logger.debug('Working dir: ' + tmpdir)
        os.chdir(tmpdir)

        kernel = self._preparekernel(kernel=self.kernel, tmpdir=tmpdir)

        initrd = self._prepareinitrd(initrd=self.initrd, tmpdir=tmpdir)

        self._unpackinitrd(initrd=initrd, tmpdir=tmpdir)

        self._setuplatecommand(tmpdir=tmpdir)

        self._setuppreseed(tmpdir=tmpdir)

        self._setuplogging(tmpdir=tmpdir)

        initrd = self._repackinitrd(tmpdir=tmpdir)

        self._createdisks()

        xml = self._installxml(cmdline=self.cmdline, initrd=initrd,
                               kernel=kernel, tmpdir=tmpdir, xml=self.xml)

        self._installvm(lv=self.lv, tmpdir=tmpdir, xml=xml)

        xml = self._finalxml(tmpdir=tmpdir, xml=xml)

        self.logger.info('Setting up final VM')
        self.vm = self.lv.defineXML(ElementTree.tostring(xml.getroot()))

        if self.debug:
            self.logger.info('Leaving temp directory '
                             'because debug is enabled: ' + tmpdir)
        else:
            self.logger.info('Cleaning up temp directory')
            shutil.rmtree(tmpdir)
        return True

    def _start(self):
        """
        Start the machine.
        """
        self.logger.info('Starting CustomVM')
        self.vm.create()
        self.logger.info('Waiting ' + str(self.boottimeout) +
                         ' seconds to allow machine to boot')
        self.sshpoll(self.boottimeout, 3)
        self.active = True

    def destroy(self, *args, **kw):
        """
        Remove the machine from libvirt and remove all the disk files.
        """
        super(CustomVM, self).destroy(*args, **kw)
        self.stop(force=True)
        self.vm.undefine()
        for disk in self.disks:
            os.unlink(disk['file'])
        shutil.rmtree(self.directory)


# See http://kennethreitz.com/blog/generate-a-random-mac-address-in-python/
def random_mac_address():
    """Returns a completely random Mac Address"""
    mac = [0x52, 0x54, 0x00, random.randint(0x00, 0xff),
                             random.randint(0x00, 0xff),
                             random.randint(0x00, 0xff)]
    return ':'.join(map(lambda x: "%02x" % x, mac))
