"""
Provide functionality to execute command with timeouts.
"""
import os
import subprocess
import signal

from utah.exceptions import UTAHException
from utah.commandstr import commandstr


class UTAHTimeout(UTAHException):
    """
    Provide a special exception to indicate an operation timed out.
    """


# Mostly pulled from utah/client/common.py
# Inspired by:
# http://stackoverflow.com/a/3326559
def timeout(timeout, command, *args, **kw):
    """
    Run command with all other arguments for up to timeout seconds, after with
    a UTAHTimeout exception is returned.
    """
    #TODO: Better support for nested timeouts.
    if command is None:
        return

    def alarm_handler(signum, frame):
        raise UTAHTimeout

    if timeout != 0:
        oldtimeout = signal.alarm(0)
        if oldtimeout > 0:
            # We're in a nested timeout.  Use the outermost one for now.
            signal.alarm(oldtimeout)
            try:
                return command(*args, **kw)
            except UTAHTimeout:
                raise UTAHTimeout(commandstr(command, *args, **kw)
                        + ' timed out after ' + str(timeout) + ' seconds')
        else:
            signal.signal(signal.SIGALRM, alarm_handler)
            signal.alarm(timeout)
            try:
                retval = command(*args, **kw)
                if timeout != 0:
                    signal.alarm(0)
                return retval
            except UTAHTimeout:
                raise UTAHTimeout(commandstr(command, *args, **kw)
                        + ' timed out after ' + str(timeout) + ' seconds')


def subprocesstimeout(timeout, *args, **kw):
    """
    Pass all arguments except timeout to subprocess.Popen and run for timeout
    seconds.
    After the timeout is reached, kill all subprocesses and return a
    UTAHTimeout exception.
    """
    if args is None:
        return

    class TimeoutAlarm(Exception):
        pass

    def alarm_handler(signum, frame):
        raise TimeoutAlarm

    if timeout != 0:
        signal.signal(signal.SIGALRM, alarm_handler)
        oldtimeout = signal.alarm(timeout)
        if oldtimeout > 0:
            signal.alarm(oldtimeout)

    try:
        p = subprocess.Popen(*args, **kw)
        while p.poll() is None:
            pass
        if timeout != 0:
            signal.alarm(0)
        return p
    except TimeoutAlarm:
        pids = [p.pid]
        # Kill p's children too.
        pids.extend(get_process_children(p.pid))

        for pid in pids:
            # process might have died before getting to this line
            # so wrap to avoid OSError: no such process
            try:
                os.kill(pid, signal.SIGKILL)
            except OSError:
                pass

        msg = (' '.join(args[0]) + ' timed out after '
               + str(timeout) + ' seconds')
        raise UTAHTimeout(msg)


def get_process_children(pid):
    """
    Find process children so they can be killed when the timeout expires.
    """
    p = subprocess.Popen('ps --no-headers -o pid --ppid %d' % pid, shell=True,
                         stdout=subprocess.PIPE, stderr=subprocess.PIPE)
    stdout, stderr = p.communicate()

    return [int(pid) for pid in stdout.split()]
